
/*
   Copyright (c) 1995-1998 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute this software for
   any purpose and without fee is hereby granted, provided that this
   copyright and permission notice appear on all copies of the
   software and supporting documentation, the name of Cisco Systems,
   Inc. not be used in advertising or publicity pertaining to
   distribution of the program without specific prior permission, and
   notice be given in supporting documentation that modification,
   copying and distribution is by permission of Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"
#include <stdio.h>
#include <errno.h>

#ifndef WITH_INCLUDED_REGEX
#ifdef HAVE_REGEX_H
#include <regex.h>
#endif
#else
#include "tac_regexp.h"
#endif

#if defined(USE_DB)
#include "db.h"
#endif /* USE_DB */

/*
   <config>         := <decl>*

   <decl>           := <top_level_decl> | <user_decl>

   <top_level_decl> := default authentication = <login_spec> |
                       accounting <acct_spec>  |
                       default authorization = permit |
                       key = <string> |
		       accesslog file = <filespec>

   <permission>     := permit | deny

   <filename>    := <string>

   <password>    := <string>

   <user_decl>   := user = <string> {
                        [ default service = [ permit | deny ] ]
                        <user_attr>*
                        <svc>*
                   }
   
   <host_decl>   :=	host = <host name or ip> <
                       key    = <string>            *** key to use for this host
					   type   = <string>
					   prompt =	<string>	        *** Username prompt
					   prompt motd = <filespec>     *** read prompt from file
					   login  = <login_spec>        *** login to use for this host
					   enable = <enable_spec>       *** enable method for this host
					   accounting = <acct_spec>     *** accounting for this host
					   permit enable group <string>	*** restrict enable to user/group
					   permit login  group <string>	*** restrict login to user/group
					   deny enable group <string>   *** deny enable if group member
					   deny login group <string>	*** deny login if group member

   <login_spec>  := file <filename> | 
		            skey | 
		            cleartext <password> |
		            des <password> |
		            pam <pam_service> |   *** if USE_PAM defined
					db <string>		  |   *** if USE_DB defined
					ldap <string>	  |   *** if USE_LDAP defined
		            nopassword
					
		*<login_spec> for host and default only allow external lists 
					

   *** Added enable_spec 6/12/02 JRM 	 
   <enable_spec> := file <filename> | 
		            cleartext <password> |
		            des <password> |
		            pam <pam_service> |      *** if USE_PAM defined
				    db <string>		  |      *** if USE_DB defined
				    ldap <string>	  |      *** if USE_LDAP defined
				    login                    *** use the same method as the login
					 
   *** Added acct_spec 6/12/02 JRM 	 
   <acct_spec>	  := file <filename> |
                     db <string>
					
   <user_attr>   :=  name     = <string> |
                     login    = <login_spec> |
		     enable   = <enable_spec> |     *** Added 6/10/02 JRM
				 [enable level <lvl>]*	|  *** Added 6/10/02 JRM
				 priv-lvl = <lvl> |  *** replace enable level 6/30/02 JRM
			 accounting = <acct_spec> |     *** Added 6/12/02 JRM
        	         member   = <string> |
        	         expires  = <string> |
                	 arap     = cleartext <string> |
	                 chap     = cleartext <string> |
	                 ms-chap  = cleartext <string> |  *** if MSCHAP defined
	                 pap      = cleartext <string> |
	                 pap      = des <string> |
	                 pap      = pam <pam_service> |	  *** if USE_PAM defined
	                 opap     = cleartext <string> |
	                 global   = cleartext <string> |
        	         msg      = <string>
		     before authorization = <string> |
                     after authorization = <string> |

   <svc>            := <svc_auth> | <cmd_auth>

   <cmd_auth>       := cmd = <string> {
                        <cmd-match>*
                    }

   <cmd-match>      := <permission> <string>

   <svc_auth>       := service = ( exec | arap | slip | ppp protocol = <string> {
                        [ default attribute = permit ]
                        <attr_value_pair>*
                    }

   <attr_value_pair> := [ optional ] <string> = <string>
   
*** Added 6/14/02 JRM
   order of precedence:  
		for accounting: user -> group -> default & host (if exists)
		my idea is that the host entries take priority
		  for login: host -> user -> group -> default
		  for enable: host -> user -> group -> $enable<n>$ -> $enable$
		this is to allow a host override to use a specific authen list
		examples: vendor routers, firewalls, etc.

*/

static char sym_buf[MAX_INPUT_LINE_LEN];	/* parse buffer */
static int sym_pos=0;						/* current place in sym_buf */
static int sym_ch;							/* current parse character */
static int sym_code;						/* parser output */
static int sym_line = 1;					/* current line number for parsing */
static FILE *cf = NULL;						/* config file pointer */
static int sym_error = 0;					/* a parsing error has occurred */
static int no_user_dflt = 0;				/* default if user doesn't exist */
static char *authen_default = NULL;			/* top level authentication default */
static int authen_default_method = 0; 		/*For method check */

static char err_buf[MAX_INPUT_LINE_LEN];	/* buffer for error msgs */
static char tmp_buf[MAX_INPUT_LINE_LEN];	/* buffer for temp use   */



#if defined(USE_DB)
	TAC_DBINFO	*db_authen_default;
#endif

/* A host definition structure. Currently unused, but when we start
   configuring host-specific information e.g. per-host keys, this is
   where it should be kept.

   The first 2 fields (name and hash) are used by the hash table
   routines to hash this structure into a table.  Do not (re)move them */

#define MAX_HOST_ACL	10
typedef struct host_acl {
	char	*name;
	void	*hash;
	int		type;
	int		permit;
	int		value;
	int		line;
} HOST_ACL;
#define HOST_ACL_PERMIT 1
#define HOST_ACL_DENY   0

struct host {
	char *name;			/* original host name spec */
	void *hash;			/* hash table next pointer */
	tac_uint32 addr;
	tac_uint32 mask;
	int line;			/* line number defined on */
	char *key;			/* host specific key */
	char *type;			/* host type - what is this? */
	char *prompt;            /* host Login prompt string */
	char *motd;				 /* host motd */
	char *enable;            /* host enable password */
	int enable_auth_method;  /* host enable type */
	char *login;             /* host login password */
	int login_auth_method;   /* host login type */
	char *acctfile;          /* host accounting file */
	int cnt_acl[2];		 	 /* login[0]/enable[1] */
	void **acltable;
#ifdef USE_DB
	TAC_DBINFO	*db_enable;		/* database info for enable */
	TAC_DBINFO	*db_login;		/* database info for login  */
	TAC_DBINFO	*db_acct;		/* database info for accounting */
#endif
};

/* A user or group definition

   The first 2 fields (name and hash) are used by the hash table
   routines to hash this structure into a table.  Move them at your
   peril */

struct user {
	char *name;			/* username */
	void *hash;			/* hash table next pointer */
	int line;			/* line number defined on */
	long flags;			/* flags field */

#define FLAG_ISUSER  1		/* this structure represents a user */
#define FLAG_ISGROUP 2		/* this structure represents a group */
#define FLAG_SEEN	4		/* for circular definition detection */

	char *full_name;		/* users full name */
	char *login;		/* Login password */
	int login_auth_method;		/* Login type */
	char *global;		/* password to use if none set */
	char *member;		/* group we are a member of */
	char *expires;		/* expiration date */
	char *arap;			/* our arap secret */
	char *pap;			/* our pap secret */
	char pap_auth_method;	/* PAP auth method */
	char *opap;			/* our outbound pap secret */
	char *chap;			/* our chap secret */
#ifdef MSCHAP
	char *mschap;		/* our mschap secret */
#endif /* MSCHAP */
	char *msg;				/* a message for this user */
	char *before_author;	/* command to run before authorization */
	char *after_author;		/* command to run after authorization */
	int svc_dflt;			/* default authorization behaviour for svc or
							 * cmd */
	NODE *svcs;				/* pointer to svc nodes */
#ifdef MAXSESS
	int maxsess;			/* Max sessions/user */
#endif /* MAXSESS */
	char *time;				/* Timestamp  */
	int enable_level;		/* enable level, default=15 */
	int enable_auth_method;		/* enable type */			
	char *enable;			/* enable scheme */
	char *acctfile;			/* per user accounting file */ 
#if defined(USE_DB)
	TAC_DBINFO	*db_acct;	/* parsed db_fields for db accounting */
	TAC_DBINFO	*db_login;	/* parsed db_fields for db login authentication */
	TAC_DBINFO	*db_enable;	/* parsed db_fields for db enable authentication */
#endif
};

typedef struct host HOST;
typedef struct user USER;

struct hostentry {
	tac_uint32	addr;
	tac_uint32	mask;
	HOST		*host;
	void		*other;
};

typedef struct hostentry HOSTENTRY;

#define DEFAULT_HOST_LIMIT 256
unsigned	cntHost = 0;
unsigned	maxHost = DEFAULT_HOST_LIMIT;
HOSTENTRY 	host_table[DEFAULT_HOST_LIMIT];

/* Only the first 2 fields (name and hash) are used by the hash table
   routines to hashh structures into a table.
*/

union hash {
	struct user u;
	struct host h;
};

typedef union hash HASH;
typedef *hashtable[HASH_TAB_SIZE];

void *grouptable[HASH_TAB_SIZE];/* Table of group declarations */
void *usertable[HASH_TAB_SIZE];	/* Table of user declarations */

static void sym_get(int flag);
#define SYM_GET_NORMAL	0
#define SYM_GET_IGNORE_EOL		1	


#ifdef __STDC__
#include <stdarg.h>		/* ANSI C, variable length args */
static void
parse_error(char *fmt,...)
#else
#include <varargs.h>		/* has 'vararg' definitions */
/* VARARGS2 */
static void
parse_error(fmt, va_alist)
char *fmt;

va_dcl				/* no terminating semi-colon */
#endif
{
	char msg[256];		/* temporary string */
	va_list ap;

#ifdef __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	vsprintf(msg, fmt, ap);
	va_end(ap);

	report(LOG_ERR, "%s on line %d", msg, sym_line);
	fprintf(stderr, "Error: %s on line %d\n", msg, sym_line);
	sym_error++;
	tac_exit(1);
}

char *
cfg_nodestring(type)
	int type;
{
	switch (type) {
	default:
		return ("unknown node type");
	case N_arg:
		return ("N_arg");
	case N_optarg:
		return ("N_optarg");
	case N_svc:
		return ("N_svc");
	case N_svc_exec:
		return ("N_svc_exec");
	case N_svc_slip:
		return ("N_svc_slip");
	case N_svc_ppp:
		return ("N_svc_ppp");
	case N_svc_arap:
		return ("N_svc_arap");
	case N_svc_cmd:
		return ("N_svc_cmd");
	case N_permit:
		return ("N_permit");
	case N_deny:
		return ("N_deny");
	}
}

static void
free_attrs(node)
NODE *node;
{
	NODE *next;

	while (node) {
	switch (node->type) {
	case N_optarg:
	case N_arg:
		if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free_cmd_match %s %s",
			   cfg_nodestring(node->type),
			   node->value);
		break;
	default:
		report(LOG_ERR, "Illegal node type %s for free_attrs", 
		   cfg_nodestring(node->type));
		return;
	}

	free(node->value);
	next = node->next;
	free(node);
	node = next;
	}
}

static void
free_cmd_matches(node)
NODE *node;
{
	NODE *next;

	while (node) {
	if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free_cmd_match %s %s",
		   cfg_nodestring(node->type),
		   node->value);

	free(node->value);	/* text */

#ifdef WITH_INCLUDED_REGEX

	free(node->value1);	/* tac_regexp compiled text */

#else /* WITH_INCLUDED_REGEX */

	regfree((regex_t *) node->value1);	/* POSIX regex compiled text */

#endif /* WITH_INCLUDED_REGEX */

	next = node->next;
	free(node);
	node = next;
	}
}

static void
free_svcs(node)
NODE *node;
{
	NODE *next;

	while (node) {

	switch (node->type) {
	case N_svc_cmd:
		if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free %s %s",
			   cfg_nodestring(node->type), node->value);
		free(node->value);	/* cmd name */
		free_cmd_matches(node->value1);
		next = node->next;
		free(node);
		node = next;
		continue;

	case N_svc:
	case N_svc_ppp:
		free(node->value1);
		/* FALL-THROUGH */
	case N_svc_exec:
	case N_svc_arap:
	case N_svc_slip:
		if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free %s", cfg_nodestring(node->type));
		free_attrs(node->value);
		next = node->next;
		free(node);
		node = next;
		continue;

	default:
		report(LOG_ERR, "Illegal node type %d for free_svcs", node->type);
		return;
	}
	}
}

static void
free_userstruct(user)
USER *user;
{
	if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "free %s %s",
			   (user->flags & FLAG_ISUSER) ? "user" : "group",
			   user->name);

	if (user->name)	{
		free(user->name);
		user->name = NULL;
	}
	if (user->full_name) {
		free(user->full_name);
		user->full_name = NULL;
	}
	if (user->login) {
		free(user->login);
		user->login = NULL;
	}
	if (user->member) {
		free(user->member);
		user->member = NULL;
	}
	if (user->expires) {
		free(user->expires);
		user->expires = NULL;
	}
	if (user->time)	{
		free(user->time);
		user->time = NULL;
	}
	if (user->arap) {
		free(user->arap);
		user->arap = NULL;
	}
	if (user->chap)	{
		free(user->chap);
		user->chap = NULL;
	}
#ifdef MSCHAP
	if (user->mschap) {
		free(user->mschap);
		user->mschap = NULL;
	}
#endif /* MSCHAP */
	if (user->pap) {
		free(user->pap);
		user->pap = NULL;
	}
	if (user->opap) {
		free(user->opap);
		user->opap = NULL;
	}
	if (user->global) {
		free(user->global);
		user->global = NULL;
	}
	if (user->msg) {
		free(user->msg);
		user->msg = NULL;
	}
	if (user->before_author) {
		free(user->before_author);
		user->before_author = NULL;
	}
	if (user->after_author) {
		free(user->after_author);
		user->after_author = NULL;
	}
	if (user->enable) {
		free(user->enable);
		user->enable = NULL;
	}
	if (user->acctfile) {
		free(user->acctfile);
		user->acctfile = NULL;
	}

#if defined(USE_DB)
	if (user->db_acct) {
		free(user->db_acct);
		user->db_acct = NULL;
	}
	if (user->db_login) {
		free(user->db_login);
		user->db_login = NULL;
	}
	if (user->db_enable) {
		free(user->db_enable);
		user->db_enable = NULL;
	}
#endif
	free_svcs(user->svcs);
}

static void
free_hoststruct(host)
HOST *host;
{
	void **acl;
	HOST_ACL	*hacl, *next_hacl;
	int cnt, idx;

	if (debug & DEBUG_CLEAN_FLAG)
		report(LOG_DEBUG, "host: free entry for '%s'", host->name);

	if (host->name)	{
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free name");

		free(host->name);
		host->name = NULL;
	}
	
	if (host->key) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free key");
		free(host->key);
		host->key = NULL;
	}
	
	if (host->type)	{
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free type");
		free(host->type);
		host->type = NULL;
	}

	if (host->prompt) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free prompt");
		free(host->prompt);
		host->prompt = NULL;
	}
	
	if (host->enable) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free enable");
		free(host->enable);
		host->enable = NULL;
	}

	if (host->login) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free login");
		free(host->login);
		host->login = NULL;
	}

#ifdef USE_DB

	if (host->db_enable) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free db_enable");
		free(host->db_enable);
		host->db_enable = NULL;
	}


	if (host->db_login) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free db_login");
		free(host->db_login);
		host->db_login = NULL;
	}


	if (host->db_acct) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free db_acct");
		free(host->db_acct);
		host->db_acct = NULL;
	}

#endif

	if (host->acctfile) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free acctfile");
		free(host->acctfile);
		host->acctfile = NULL;
	}

	if (host->acltable) {
		if (debug & DEBUG_CLEAN_FLAG)
			report(LOG_DEBUG, "host: free acltable");
		for (cnt = 0, acl = host->acltable; cnt < HASH_TAB_SIZE; cnt ++, acl++) {
			hacl = (HOST_ACL *) *acl;
			while (hacl) {
				if (hacl->name) {
					if (debug & DEBUG_CLEAN_FLAG)
						report(LOG_DEBUG, "host: free %s[%d]", hacl->name, cnt);
					free(hacl->name);
				}
				next_hacl = hacl->hash;
				free(hacl);
				hacl = next_hacl;
			}
			*acl = NULL;
		}
		free(host->acltable);
	}

}

/*
 * Exported routines
 */

/* Free all allocated structures preparatory to re-reading the config file */
void
cfg_clean_config()
{
	int i;
	USER *entry, *next;
	HOST *host;

	if (authen_default) {
		free(authen_default);
		authen_default = NULL;
	}
   
	if (authen_default_method) {
		authen_default_method = 0;
	}

	if (session.key) {
		free(session.key);
		session.key = NULL;
	}

	if (session.prompt) {
		free(session.prompt);
		session.prompt = NULL;
	}

	if (session.acctfile) {
		free(session.acctfile);
		session.acctfile = NULL;
	}

	if (session.accesslog_file) {
		free(session.accesslog_file);
		session.accesslog_file = NULL;
	}
	
#if defined(USE_DB)
	if (db_authen_default) {
		free(db_authen_default);
		db_authen_default = NULL;
	}

	if (session.db_acct) {
		free(session.db_acct);
		session.db_acct = NULL;
	}

	if (session.db_default) {
		free(session.db_default);
		session.db_default = NULL;
	}
#endif

	/* clean the hosttable */

	for (i = 0; i < cntHost; i++) {
		if (host = host_table[i].host) {
			free_hoststruct(host);
			free(host);
		}
		bzero(&(host_table[i]), sizeof(HOSTENTRY));
	}
	cntHost = 0;

	/* the grouptable */
	for (i = 0; i < HASH_TAB_SIZE; i++) {
		entry = (USER *) grouptable[i];
		while (entry) {
			next = entry->hash;
			free_userstruct(entry);
			free(entry);
			entry = next;
		}
	grouptable[i] = NULL;
	}

	/* the usertable */
	for (i = 0; i < HASH_TAB_SIZE; i++) {
		entry = (USER *) usertable[i];
		while (entry) {
			next = entry->hash;
			free_userstruct(entry);
			free(entry);
			entry = next;
		}
	usertable[i] = NULL;
	}
}

static int
parse_permission()
{
	int symbol = sym_code;

	if (sym_code != S_permit && sym_code != S_deny) {
	   parse_error("expecting permit or deny but found '%s'", sym_buf);
	return (0);
	}
	sym_get(0);

	return (symbol);
}

static int non_string[] = { S_openbra, S_closebra, S_optional, S_unknown, S_comma,
						    S_separator, S_eof, S_eol, S_openparen, S_closeparen, 0 };

static int
parse(symbol)
int symbol;
{
	int idx;
	int test_sym;

	if (symbol == S_string) {
		idx = 0;
		while (test_sym = non_string[idx++]) {
			if (test_sym == sym_code) {
				parse_error("expecting string but found '%s'", sym_buf);
				return (1);
			}
		}
	} else if (sym_code != symbol) {
		parse_error("expecting '%s' but found '%s'",
			codestring(symbol), sym_buf);
		return (1);
	}
	sym_get(0);
	return (0);
}

static int
parse_opt_svc_default()
{
	if (sym_code != S_default) {
	return (0);
	}

	parse(S_default);
	parse(S_svc);
	parse(S_separator);
	if (sym_code == S_permit) {
	parse(S_permit);
	return (S_permit);
	}
	parse(S_deny);
	return (S_deny);
}

static int
parse_opt_attr_default()
{
	if (sym_code != S_default)
	return (S_deny);

	parse(S_default);
	parse(S_attr);
	parse(S_separator);
	parse(S_permit);
	return (S_permit);
}

static int parse_user();
static int parse_host();

static void
 rch();

/*
   Parse lines in the config file, creating data structures
   Return 1 on error, otherwise 0 */

static int
parse_decls()
{
	no_user_dflt = 0; /* default if user doesn't exist */

	sym_code = 0;
	rch();

	bzero(grouptable, sizeof(grouptable));
	bzero(usertable, sizeof(usertable));
	bzero(host_table, sizeof(host_table)); 

	sym_get(0);

	/* Top level of parser */
	while (1) {

		switch (sym_code) {
		case S_eof:
			return (0);

		case S_eol:
			sym_get(0);
			continue;
	
		case S_accounting:
			sym_get(0);
	
			switch (sym_code) {
			case S_file:
				/* accounting file  */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (session.acctfile) {
					free(session.acctfile);
				}
				session.acctfile = tac_strdup(sym_buf);
				break;
	
#ifdef USE_DB
			case S_db:
				/* accounting database */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (session.db_acct) {
					free(session.db_acct);
				}
				if (NULL == (session.db_acct = db_parse_url(sym_buf))) {
					parse_error("parsing db url '%s'", sym_buf);
				}
				break;
#endif
	
			default:
				strcpy(err_buf, "'file'");
#ifdef USE_DB
				strcat(err_buf, "or 'db'");
#endif
				parse_error("expecting %s but found '%s'", 
							err_buf, sym_buf);
				return(1);
			} /* switch */

			sym_get(0);
			continue;
	
#ifdef USE_DB	
		case S_db_accounting:
			sym_get(0);
			parse_error("db_accounting = <string> is now deprecated.Please use accounting db instead.");
/**
			if (sym_code == S_separator) {
				sym_get(0);
			}
			if (session.db_acct) 
				free(session.db_acct);
			if (NULL == (session.db_acct = db_parse_url(sym_buf))) {
				parse_error("parsing db url '%s'", sym_buf);
			}
**/
			sym_get(0);
			continue;
#endif
	
		case S_default:
			sym_get(0);
			switch (sym_code) {
			default:
				parse_error(
					"Expecting default authorization/authentication");
				break;
	
			case S_authentication:
				if (authen_default) {
					parse_error("Multiply defined authentication default");
					break;
				}
				parse(S_authentication);
				parse(S_separator);
	
				authen_default_method = sym_code;

				switch(sym_code) {
				
				case S_unix:   /* use the unix getpw* functions */
				case S_skey:
					authen_default = tac_strdup(sym_buf);
					break;
					
				case S_file:
					sym_get(0);
					authen_default = tac_strdup(sym_buf);
					if ((STREQ(sym_buf, "/etc/passwd") || STREQ(sym_buf,"/etc/shadow"))) {
						/* use the getpw* functions instead of direct file reads */
						authen_default_method = S_unix;
					}
					break;

#ifdef USE_LDAP
				case S_ldap:
#endif
#ifdef USE_PAM
				case S_pam:
#endif
					sym_get(0);
					authen_default = tac_strdup(sym_buf);
					break;
#ifdef USE_DB
				case S_db:
					sym_get(0);
					authen_default = tac_strdup(sym_buf);
					db_authen_default = db_parse_url(sym_buf);
					if (!db_authen_default) {
						parse_error("parsing db url '%s'", sym_buf);
					}
					break;
#endif
	
				default:
					authen_default_method = 0;

					parse_error("expecting default_method keyword after " \
								"'default authentication = '");
					break;
				} /* switch */
				sym_get(0);
				continue;
	
			case S_authorization:
				parse(S_authorization);
				parse(S_separator);
				parse(S_permit);
				no_user_dflt = S_permit;
				report(LOG_INFO, "default authorization = permit " \
					   "is now deprecated. Please use user = DEFAULT instead");
				continue;

#ifdef USE_DB
			case S_db:
				parse(S_db);
				parse(S_separator);
				report(LOG_INFO, "parced default db: %s", sym_buf);
				if (NULL==(session.db_default = db_parse_url(sym_buf)))
					parse_error("parcing default db: %s",sym_buf);
				sym_get(0);
				continue;
#endif
			} /* switch */
	
		case S_key:
			/* Process a key declaration. */
			sym_get(0);
			parse(S_separator);
			if (session.key) {
				parse_error(
					"multiply defined key on lines %d and",session.keyline);
				break;
			}
			session.key = tac_strdup(sym_buf);
			session.keyline = sym_line;
			sym_get(0);
			continue;
		
		case S_prompt:
			sym_get(0);
			parse(S_separator);
			if (session.prompt) {
				parse_error("multiply defined prompt on lines %d and %d", session.keyline, sym_line);
				return (1);
			}
			session.prompt = tac_strdup(sym_buf);
			session.keyline = sym_line;
			sym_get(0);
			continue;

		case S_host:
			parse_host();
			continue;
		
		case S_user:
		case S_group:
			parse_user();
			continue;

		case S_accesslog:
			sym_get(0);
			switch (sym_code) {
			case S_file:
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (session.accesslog_file) {
					free(session.accesslog_file);
				}
				session.accesslog_file = tac_strdup(sym_buf);
				break;

			default:
				parse_error("expecting 'file' but found '%s'", sym_buf);
				return(1);
			}
			sym_get(0);
			continue;

	
		default:
			parse_error("Unrecognized token '%s'", sym_buf );
			return (1);
		} /* switch */
	} /* while */
	return( 0);
}



static NODE *parse_svcs();

/* Assign a value to a field. Issue an error message and return 1 if
   it's already been assigned. This is a macro because I was sick of
   repeating the same code fragment over and over */

#define ASSIGN(_field) \
sym_get(0); parse(S_separator); if (_field) { \
	parse_error("Duplicate value for %s %s and %s", \
			codestring(sym_code), _field, sym_buf); \
	} else { _field = tac_strdup(sym_buf); }

static HOSTENTRY *
insert_host_entry(HOST *host)
{
	HOSTENTRY *he;
	unsigned	shift, idxInsert = 0;

	if (!host) {
		return(NULL);
	}

	if (!cntHost) {			/* first entry is easy, just drop it in */
		idxInsert = cntHost;
	} else {
		/* look for the best fit */
		
		for (idxInsert = 0; idxInsert < cntHost; idxInsert++) {
		   if (host->mask > host_table[idxInsert].mask) {
			   break;
		   } else if (host->mask == host_table[idxInsert].mask &&
					  host->addr == host_table[idxInsert].addr) {
			   if (debug & DEBUG_PARSE_FLAG) {
				   report(LOG_ERR, "Duplicate host definition: " \
						  "definition '%s' on line %d is the same as " \
						  "'%s' on line %d",
						  host->name, host->line,
						  host_table[idxInsert].host->name,
						  host_table[idxInsert].host->line );

			   }
			   return (NULL);
		   }
		}
		/* calc shift */
		shift = cntHost - idxInsert;

		if (shift) { 
			/* move entries down */
			memmove(&(host_table[idxInsert+1]), 
					&(host_table[idxInsert]), 
					sizeof(HOSTENTRY) * shift);
		}
	}

	if (debug & DEBUG_PARSE_FLAG || parse_only) {
		report(LOG_DEBUG, "Inserting host '%s' at index %d",
				   host->name, idxInsert);
		}

	if (NULL != (he = &(host_table[idxInsert]))) {
		he->addr = host->addr;
		he->mask = host->mask;
		he->host = host;
		cntHost++;
	}
	return(he);
}

static void
dump_host_table()
{
	char szHost[16], szMask[16];
	unsigned idx;
	HOST	*host;

	report(LOG_DEBUG, "Dumping of %d entries in host table",
		   cntHost);

	for (idx = 0; idx < cntHost; idx++) {
		strcpy(szHost, cfg_addr_as_string(host_table[idx].addr));
		strcpy(szMask, cfg_addr_as_string(host_table[idx].mask));
		host = host_table[idx].host;
		report(LOG_DEBUG, "%d: host '%s' as %s %s",
			   idx, host->name, szHost, szMask);
	}

}

static char *read_motd(char *filename)
{
	FILE	*motd;
	long	filelen;
	char 	*text;

	if (!filename) {
		return NULL;
	}

	if (NULL == (motd = fopen(filename, "r"))) {
		if (debug & DEBUG_PARSE_FLAG) {
			report(LOG_ERR, "read_motd: Unable to open motd file: '%s': %s", filename, strerror(errno));
		}
		return NULL;
	}
	fseek( motd, 0, SEEK_END);
	filelen = ftell(motd);
	if (debug & DEBUG_PARSE_FLAG) {
		report(LOG_DEBUG, "read_motd: file is %ld bytes long", filelen);
	}
	text = tac_malloc(filelen + 32);
	bzero(text, filelen + 32);
	fseek( motd, 0, SEEK_SET);
	fread(text, filelen, sizeof(char), motd);
	fclose(motd);

	return( text);
}

static int
parse_host_acl(HOST *host, int permit_type)
{
	int auth_type;
	HOST_ACL	*hacl;
	int acl_value = 0;
	int bMoreUsers = 1;  /* there should be at least one user to start off this acl */

	if (!host) {  
		/* quick check to see if we have to die */
		parse_error("parse_host_acl called without host");
		return(0);
	}

	sym_get(0);
	auth_type = sym_code;

	if (auth_type != S_login && auth_type != S_enable) {
		parse_error("Expecting 'login' or 'enable' following keyword %s", 
					codestring(permit_type));
		return(0);
	}

	if (!host->acltable) {
		if (NULL == (host->acltable = (void **) calloc(sizeof(char *), HASH_TAB_SIZE))) {
			report(LOG_ERR, "unable to allocate memory for host acl table");
			tac_exit(1);
		}
	}

	sym_get(0);

	/* check to see if global level specified */
	if (sym_code == S_openparen) {
		if (auth_type != S_enable) { /* but only on enable */
			parse_error ("Level only valid on enable acl");
			return(0);
		}

		sym_get(0);
		acl_value = atoi(sym_buf);
		if (acl_value < TAC_PLUS_PRIV_LVL_MIN ||
			acl_value > TAC_PLUS_PRIV_LVL_MAX) {
			parse_error("priv-lvl '%s' is invalid",sym_buf);
			return(0);
		}
		sym_get(0);
		parse(S_closeparen);
	}

	while (bMoreUsers) {
		hacl = (HOST_ACL *) tac_malloc(sizeof(HOST_ACL));
		hacl->type = auth_type;
		host->cnt_acl[hacl->type == S_login ? 0 : 1]++;
		hacl->name = tac_strdup(sym_buf);
		hacl->line = sym_line;
		hacl->permit = (permit_type == S_permit);
		hacl->value = acl_value;  /* first set the global value */

		sym_get(0);  
		if (sym_code == S_openparen) {
			/* we have a level set for the user */
			if (auth_type != S_enable) { /* but only on enable */
				parse_error ("Level only valid on enable acl");
				break;
			}
			sym_get(0);
			hacl->value = atoi(sym_buf);
			if (hacl->value < TAC_PLUS_PRIV_LVL_MIN ||
				hacl->value > TAC_PLUS_PRIV_LVL_MAX) {
				parse_error("priv-lvl '%s' is invalid",sym_buf);
				break;
			}
			sym_get(0);

			parse(S_closeparen);
		}

		if (hacl->value) {
			if (debug & DEBUG_PARSE_FLAG) {
				report(LOG_DEBUG, "priv-lvl will be set to '%d' for user '%s' on host '%s'",
					   hacl->value, hacl->name, host->name);
			}
		}

		if (debug & DEBUG_PARSE_FLAG) {
			report(LOG_DEBUG, "adding acl '%s %s %s' to host '%s'",
				   codestring(hacl->type),
				   hacl->permit ?  "permit" : "deny",
				   hacl->name, host->name);
		}

		if (NULL != hash_add_entry(host->acltable, hacl)) {
			/* we have a user or group specified more than once */
			parse_error("Duplicate name '%s' on host ACL", hacl->name);
			return(0);
		}

		/* here we are positioned on either the list separator (comma) or the next command */

		if (bMoreUsers = (sym_code == S_comma)) {
			/* a comma means more users */
			sym_get(0);
		}
	} /* while to get the whole list */

	return(1);

}


static int
parse_host()
{
	HOST *host = (HOST *) tac_malloc(sizeof(HOST));
	struct hostent *hp;
	HOSTENTRY	*he;
	char szTmp1[32], szTmp2[32];

	sym_get(0);
	/* the = following host is optional */
	if (sym_code == S_separator) {
		sym_get(0);
	}
	host->name = tac_strdup(sym_buf);
	host->line = sym_line;

	if (!parse_host_name(host->name, &(host->addr), &(host->mask))) {
		hp = gethostbyname(host->name);
		if (!hp) {
			parse_error("Host name %s is invalid or cannot be resolved", host->name);
			return(1);
		}
		host->mask = 0xFFFFFFFF;
		memcpy(&(host->addr), hp->h_addr, sizeof(host->addr));
	}

	if (cntHost >= maxHost) {
		parse_error("Maximum number of host entries(%d) exceeded.", maxHost);
		return(1);
	}

	if (NULL == (he = insert_host_entry(host))) {
		parse_error("adding host entry");
		return(1);
	}

	if (debug & DEBUG_PARSE_FLAG || parse_only) {
		strcpy(szTmp1, cfg_addr_as_string(host->addr));
		strcpy(szTmp2, cfg_addr_as_string(host->mask));
		report(LOG_DEBUG, "Adding host '%s' as %s %s",
			   sym_buf, szTmp1, szTmp2);
	}

	sym_get(1);
	parse(S_openbra);
	
	while (1) {
		switch (sym_code) {
		case S_eof:
			return (0);

		case S_eol:
			sym_get(0);
			continue;

		case S_key:
			ASSIGN(host->key);
			sym_get(0);
			continue;
	
		case S_type:
			ASSIGN(host->type);
			sym_get(0);
			continue;
	
		case S_prompt:
			sym_get(0);
			if (sym_code == S_separator) {
				sym_get(0);
				if (host->prompt) {
					parse_error("Duplicate value for prompt %s and %s",
								host->prompt, sym_buf);
				} else {
					host->prompt = tac_strdup(sym_buf);
				}
			} else if(sym_code == S_motd) {
				sym_get(0);
				parse(S_separator);
				if (host->motd) {
					parse_error("Duplicate value for prompt motd");
				} else {
					host->motd = read_motd(sym_buf);
				}
			} else {
				parse_error("expecting 'motd' or '=' but found '%s'", sym_code);
				return(1);
			}
			sym_get(0);
			continue;	

		case S_login:
			if (host->login) {
				parse_error("Duplicate value for %s %s and %s",
					codestring(sym_code), host->login,
					sym_buf);
			}
			sym_get(0);
			parse(S_separator);

			host->login_auth_method = sym_code;

			/* use 'list based' authentication schemes for login */

			switch(sym_code) {
	
			case S_unix:   /* use the unix getpw* functions */
			case S_skey:
				host->login = tac_strdup(sym_buf);
				break;

			case S_file:
				sym_get(0);
				host->login = tac_strdup(sym_buf);
				if ((STREQ(sym_buf, "/etc/passwd") || STREQ(sym_buf,"/etc/shadow"))) {
					/* use the getpw* functions instead of direct file reads */
					host->login_auth_method = S_unix;
				}
				break;

#ifdef USE_LDAP
			case S_ldap:
#endif
#ifdef USE_PAM	
			case S_pam:	
#endif /* USE_PAM */
				sym_get(0);
				host->login = tac_strdup(sym_buf);
				break;
		
#ifdef USE_DB
			case S_db:
				sym_get(0);
				host->login = tac_strdup(sym_buf);
				if (NULL == (host->db_login = db_parse_url(sym_buf))) {
						parse_error("parsing db url '%s'", sym_buf);
				}
#endif
	
			default:
				host->login_auth_method = 0;

				strcpy(err_buf, "'file', 'skey', 'unix'");
#ifdef USE_PAM
				strcat(err_buf, ", 'pam'");
#endif
#ifdef USE_DB
				strcat(err_buf, ", 'db'");
#endif
				parse_error("expecting keyword of %s after 'login ='", 
							err_buf, sym_line);
			} /* switch sym_code */
			sym_get(0);
			continue;

		case S_enable:
			if (host->enable) {
				parse_error("Duplicate value for %s %s and %s",
					codestring(sym_code), host->enable, sym_buf);
			}
			sym_get(0);
			parse(S_separator);

			host->enable_auth_method = sym_code;

			switch(sym_code) {

			case S_unix:   /* use the unix getpw* functions */
			case S_login:  /* use the login method as the enable method */
				host->enable = tac_strdup(sym_buf);
				break;
		
			case S_file:
				sym_get(0);
				host->enable = tac_strdup(sym_buf);
				if ((STREQ(sym_buf, "/etc/passwd") || STREQ(sym_buf,"/etc/shadow"))) {
					/* use the getpw* functions instead of direct file reads */
					host->enable_auth_method = S_unix;
				}
				break;

			case S_cleartext:
			case S_des:
#ifdef USE_LDAP
			case S_ldap:
#endif
#ifdef USE_PAM	
			case S_pam:	
#endif /* USE_PAM */
				sym_get(0);
				host->enable = tac_strdup(sym_buf); 
				break;
		
#ifdef USE_DB
			case S_db:
				sym_get(0);
				host->enable = tac_strdup(sym_buf); 
				if (host->enable_auth_method == S_db) {
					if (NULL == (host->db_enable = db_parse_url(sym_buf))) {
						parse_error("parsing db url '%s'", sym_buf);
					}
				}
				break;
#endif

			default:
				host->enable_auth_method = 0;
				strcpy(err_buf, "'file', 'unix', 'cleartext', 'skey', 'des', 'login'");
#ifdef USE_PAM
				strcat(err_buf, ", 'pam'");
#endif
#ifdef USE_DB
				strcat(err_buf, ", 'db'");
#endif
				parse_error("expecting keyword of %s after 'enable ='", err_buf);
			}
			sym_get(0);
			continue;	
	
		case S_accounting:
			sym_get(0);
			switch (sym_code) {
			case S_file:
				/* accounting file */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (host->acctfile) {
					free(host->acctfile);
				}
				host->acctfile = tac_strdup(sym_buf);
				break;
#ifdef USE_DB
			case S_db:
				/* accounting database */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (host->db_acct) {
					free(host->db_acct);
				}
				if (NULL == (host->db_acct = db_parse_url(sym_buf))) {
					parse_error("parsing db url '%s'", sym_buf);
				}
				break;
#endif
			default:
				strcpy(err_buf, "'file'");
#ifdef USE_DB
				strcat(err_buf, "or 'db'");
#endif
				parse_error("expecting %s but found '%s'", 
							err_buf, sym_buf);
				break;

			}
			sym_get(0);
			continue;

		case S_deny:
		case S_permit:
			parse_host_acl(host, sym_code);
			continue;

		case S_closebra:
			parse(S_closebra);

			/* combine prompts (if both specified) */
			if (host->motd) {
				int		newlen;
				char	*newtext;
				char	*prompt = "";

				if (host->prompt) {	 /* they need to be combined */
					prompt = host->prompt;
				}
				newlen = strlen(host->motd) + strlen(prompt) + 16;	/* combined lengths + padding */
                newtext = tac_malloc(newlen);
				strcpy(newtext, host->motd);
				if (*prompt) {
					strcat(newtext, "\n");
					strcat(newtext, prompt);
				}
				if (host->prompt) {
					free(host->prompt);
				}
			   	host->prompt = newtext;
			   	free(host->motd);
			   	host->motd = NULL;

			}

			return (0);

		default:
			parse_error("Unrecognized keyword %s for host %s",
						sym_buf, host->name);

			return (0);
		} /* switch */
	} /* while */

} /* finish parse_host */


static int
parse_user()
{
	USER *n;
	int isuser;
	USER *user = (USER *) tac_malloc(sizeof(USER));
	int save_sym, lvl;
	char **fieldp = NULL;
#ifdef MAXSESS
	int maxsess;
#endif

	bzero(user, sizeof(USER));

	isuser = (sym_code == S_user);

	sym_get(0);
	/* the = following user/group is optional */
	if (sym_code == S_separator) {
		sym_get(0);
	}
	user->name = tac_strdup(sym_buf);
	user->line = sym_line;

	if (isuser) {
		user->flags |= FLAG_ISUSER;
		n = hash_add_entry(usertable, (void *) user);
	} else {
		user->flags |= FLAG_ISGROUP;
		n = hash_add_entry(grouptable, (void *) user);
	}

	if (n) {
		parse_error("multiply defined %s %s on lines %d and",
			isuser ? "user" : "group",
			user->name, n->line);
		return (1);
	}
	sym_get(1);
	parse(S_openbra);

	/* Is the default deny for svcs or cmds to be overridden? */
	user->svc_dflt = parse_opt_svc_default();

	while (1) {
		if (debug && parse_only) {
			fprintf(stderr, "user: parse code: %d\n", sym_code);
		}
		switch (sym_code) {
		case S_eof:
			return (0);

		case S_eol:
			sym_get(0);
			continue;

		case S_time:
			ASSIGN(user->time);
			sym_get(0); 
			continue;

		case S_before:
			sym_get(0);
			parse(S_authorization);
			if (user->before_author)
				free(user->before_author);
			user->before_author = tac_strdup(sym_buf);
			sym_get(0);
			continue;

		case S_after:
			sym_get(0);
			parse(S_authorization);
			if (user->after_author)
				free(user->after_author);
			user->after_author = tac_strdup(sym_buf);
			sym_get(0);
			continue;

		case S_svc:
		case S_cmd:
			
			if (user->svcs) {   
			/* 
			 * Already parsed some services/commands. Thanks to Gabor Kiss
			 * who found this bug.
			 */
				NODE *p;
				for (p=user->svcs; p->next; p=p->next) 
					/* NULL STMT */;
				p->next = parse_svcs();
			} else {
				user->svcs = parse_svcs();
			}
			continue;

		case S_login:
			if (user->login) {
				parse_error("Duplicate value for %s %s and %s",
					codestring(sym_code), user->login, sym_buf);
			}
			sym_get(0);
			parse(S_separator);

			user->login_auth_method = sym_code;

			switch(sym_code) {
	
			case S_unix:   /* use the unix getpw* functions */
			case S_nopasswd:
			case S_skey:
				user->login = tac_strdup(sym_buf);
				break;

			case S_file:
				sym_get(0);
				user->login = tac_strdup(sym_buf);
				if ((STREQ(sym_buf, "/etc/passwd") || STREQ(sym_buf,"/etc/shadow"))) {
					/* use the getpw* functions instead of direct file reads */
					user->login_auth_method = S_unix;
				}
				break;

			case S_cleartext:
			case S_des:
#ifdef USE_LDAP
			case S_ldap:
#endif
#ifdef USE_PAM	
			case S_pam:	
#endif /* USE_PAM */
				sym_get(0);
				user->login = tac_strdup(sym_buf);
				break;
#ifdef USE_DB
			case S_db:
				sym_get(0);
				user->login = tac_strdup(sym_buf);
				if (NULL == (user->db_login = db_parse_url(sym_buf))) {
					parse_error("parsing db url '%s'", sym_buf);
				}
				break;
#endif /* USE USE_DB */
	
			default:
				user->login_auth_method = 0;

				strcpy(err_buf, "'file', 'unix, 'cleartext', 'nopassword', 'skey', 'des'");
#ifdef USE_PAM
				strcat(err_buf, ", 'pam'");
#endif
#ifdef USE_DB
				strcat(err_buf, ", 'db'");
#endif
				parse_error("expecting keyword of %s after 'login ='", 
							err_buf, sym_line);
			} /* switch sym_code */

			sym_get(0);
			continue;

		case S_priv_lvl:
			sym_get(0);
			parse(S_separator);

			lvl = atoi(sym_buf);
			if (lvl < TAC_PLUS_PRIV_LVL_MIN || lvl > TAC_PLUS_PRIV_LVL_MAX) {
				parse_error("priv-lvl '%s' is invalid",sym_buf);
			}
			/* use the highest level specified */
			if (lvl > user->enable_level) {
				user->enable_level = lvl;
			}
			if (debug && parse_only) {
				fprintf(stderr, "user: enable set to %d\n", user->enable_level);
			}
			sym_get(0);
			continue;

		case S_enable:

			sym_get(0);

			if (user->enable) {
				free(user->enable);
				user->enable = NULL;
			}

			/* get enable authentication method */
			parse(S_separator);

			user->enable_auth_method = sym_code;

			switch(sym_code) {
	
			case S_login:  /* use the login method as the enable method - Added 6/10/02 JRM */
			case S_unix:   /* use the unix getpw* functions */
			case S_skey:
				user->enable = tac_strdup(sym_buf);
				break;

			case S_file:
				sym_get(0);
				user->enable = tac_strdup(sym_buf);
				if ((STREQ(sym_buf, "/etc/passwd") || STREQ(sym_buf,"/etc/shadow"))) {
					/* use the getpw* functions instead of direct file reads */
					user->enable_auth_method = S_unix;
				}
				break;

			case S_cleartext:
			case S_des:
#ifdef USE_LDAP
			case S_ldap:
#endif
#ifdef USE_PAM	
			case S_pam:	
#endif /* USE_PAM */ 
				sym_get(0);
				user->enable = tac_strdup(sym_buf);
				break;
#ifdef USE_DB
			case S_db:
				sym_get(0);
				user->enable = tac_strdup(sym_buf);
				if (NULL == (user->db_enable = db_parse_url(sym_buf))) {
					parse_error("parsing db url '%s'", sym_buf);
				}
				break;
#endif

			default:
				user->enable_auth_method = 0;
				strcpy(err_buf, "'file', 'cleartext', 'skey', 'des'");
#ifdef USE_PAM
				strcat(err_buf, ", 'pam'");
#endif
#ifdef USE_DB
				strcat(err_buf, ", 'db'");
#endif
				parse_error("expecting keyword of %s after 'enable ='", err_buf);
			} /* switch sym_code */

			sym_get(0);
			continue;

		case S_pap:
			if (user->pap) {
				parse_error("Duplicate value for %s %s and %s",
					codestring(sym_code), user->pap, sym_buf);
			}
			sym_get(0);
			parse(S_separator);
	
			user->pap_auth_method = sym_code;
			switch(sym_code) {

			case S_cleartext:
			case S_des:
#ifdef USE_PAM	
			case S_pam:
#endif /*USE_PAM */				
				sym_get(0);
				user->pap = tac_strdup(sym_buf);
				break;	

			default:
				strcpy(err_buf, "'file', 'cleartext', 'nopassword', 'skey', 'des'");
#ifdef USE_PAM
				strcat(err_buf, ", 'pam'");
#endif
				parse_error("expecting keyword of %s after 'pam ='", err_buf);
			}
			sym_get(0);
			continue;

		case S_name:
			ASSIGN(user->full_name);
			sym_get(0);
			continue;

		case S_member:
			ASSIGN(user->member);
			sym_get(0);
			continue;

		case S_expires:
			ASSIGN(user->expires);
			sym_get(0);
			continue;
		
		case S_message:
			ASSIGN(user->msg);
			sym_get(0);
			continue;

		case S_arap:
		case S_chap:
#ifdef MSCHAP
		case S_mschap:
#endif /* MSCHAP */
		case S_opap:
		case S_global:
			save_sym = sym_code;
			sym_get(0); 
			parse(S_separator);
			strcpy(tmp_buf, sym_buf);
			parse(S_cleartext);
			strcat(tmp_buf, " ");
			strcat(tmp_buf, sym_buf);

			if (save_sym == S_arap) 	fieldp = &user->arap;
			if (save_sym == S_chap) 	fieldp = &user->chap;
#ifdef MSCHAP
			if (save_sym == S_mschap) 	fieldp = &user->mschap;
#endif /* MSCHAP */
			if (save_sym == S_pap)		fieldp = &user->pap;
			if (save_sym == S_opap)	    fieldp = &user->opap;
			if (save_sym == S_global)	fieldp = &user->global;

			if (*fieldp) {
				parse_error("Duplicate value for %s %s and %s",
					codestring(save_sym), *fieldp, sym_buf);
			}
			*fieldp = tac_strdup(tmp_buf);
			sym_get(0);
			continue;

		case S_closebra:
			parse(S_closebra);
			return (0);

#ifdef MAXSESS
		case S_maxsess:
			maxsess=0;

			sym_get(0); 
			parse(S_separator);
			/*Modified not to use scanf. maxsess, if specified must be > 0 */
			maxsess = atoi(sym_buf);
			if (!maxsess) {
				parse_error("Max sessions must be greater than 0");
				break;
			}
			if (maxsess > user->maxsess) {
				user->maxsess = maxsess;
			}
/*
			if (sscanf(sym_buf, "%d", &user->maxsess) != 1) {
				parse_error("expecting integer, found '%s'", sym_buf);
			}
*/
			sym_get(0);
			continue;
#endif /* MAXSESS */

		case S_accounting:
			sym_get(0);
			switch (sym_code) {
			case S_file:
				/* accounting file */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (user->acctfile) {
					free(user->acctfile);
				}
				user->acctfile = tac_strdup(sym_buf);
				break;
#ifdef USE_DB
			case S_db:
				/* accounting database */
				sym_get(0);
				if (sym_code == S_separator) {
					sym_get(0);
				}
				if (user->db_acct) {
					free(user->db_acct);
				}
				if (NULL == (user->db_acct = db_parse_url(sym_buf))) {
					parse_error("parsing db url '%s'", sym_buf);
				}
				break;
#endif

			default:
				strcpy(err_buf, "'file'");
#ifdef USE_DB
				strcat(err_buf, "or 'db'");
#endif
				parse_error("expecting %s but found '%s'", 
							err_buf, sym_buf);
				break;
			}
			sym_get(0);
			continue;

		default:
			if (STREQ(sym_buf, "password")) {
				fprintf(stderr,
					"\npassword = <string> is obsolete. " \
					"Use login = des <string>\n");
			}
			parse_error("Unrecognized keyword '%s' for user",sym_buf);

			return (0);
		} /* switch sym_code */
	} /* while */
}

static NODE *parse_attrs();
static NODE *parse_cmd_matches();

static NODE *
parse_svcs()
{
	NODE *result;

	switch (sym_code) {
	default:
		return (NULL);
	case S_svc:
	case S_cmd:
		break;
	}

	result = (NODE *) tac_malloc(sizeof(NODE));

	bzero(result, sizeof(NODE));
	result->line = sym_line;

	/* cmd declaration */
	if (sym_code == S_cmd) {
		parse(S_cmd);
		parse(S_separator);
		result->value = tac_strdup(sym_buf);

		sym_get(1);
		parse(S_openbra);

		result->value1 = parse_cmd_matches();
		result->type = N_svc_cmd;

		parse(S_closebra);
		result->next = parse_svcs();
		return (result);
	}

	/* svc declaration */
	parse(S_svc);
	parse(S_separator);
	switch (sym_code) {
	default:
	parse_error("expecting service type but found %s",sym_buf);
	return (NULL);

	case S_string:
	result->type = N_svc;
	/* should perhaps check that this is an allowable service name */
	result->value1 = tac_strdup(sym_buf);
	break;
	case S_exec:
	result->type = N_svc_exec;
	break;
	case S_arap:
	result->type = N_svc_arap;
	break;
	case S_slip:
	result->type = N_svc_slip;
	break;
	case S_ppp:
	result->type = N_svc_ppp;
	parse(S_ppp);
	parse(S_protocol);
	parse(S_separator);
	/* Should perhaps check that this is a known PPP protocol name */
	result->value1 = tac_strdup(sym_buf);
	break;
	}
	sym_get(1);
	parse(S_openbra);
	result->dflt = parse_opt_attr_default();
	result->value = parse_attrs();
	parse(S_closebra);
	result->next = parse_svcs();
	return (result);
}

/*  <cmd-match>	 := <permission> <string> */

static NODE *
parse_cmd_matches()
{
	NODE *result;

	if (sym_code != S_permit && sym_code != S_deny) {
		return (NULL);
	}
	result = (NODE *) tac_malloc(sizeof(NODE));

	bzero(result, sizeof(NODE));
	result->line = sym_line;

	result->type = (parse_permission() == S_permit) ? N_permit : N_deny;
	result->value = tac_strdup(sym_buf);

#ifdef WITH_INCLUDED_REGEX

	result->value1 = (void *) tac_regcomp(result->value);

#else /* WITH_INCLUDED_REGEX */

	    result->value1 = tac_malloc(sizeof(regex_t));
	    if (regcomp(result->value1, result->value /* regex */, REG_NOSUB /* cflags */)) {
		free(result->value1);
		result->value1 = NULL;
	    }

#endif /* WITH_INCLUDED_REGEX */

	if (!result->value1) {
		parse_error("in regular expression '%s'", sym_buf);
	}
	sym_get(0);

	result->next = parse_cmd_matches();

	return (result);
}

static NODE *
parse_attrs()
{
	NODE *result;
	char buf[MAX_INPUT_LINE_LEN];
	int optional = 0;

	if (sym_code == S_closebra) {
		return (NULL);
	}
	result = (NODE *) tac_malloc(sizeof(NODE));

	bzero(result, sizeof(NODE));
	result->line = sym_line;

	if (sym_code == S_optional) {
		optional++;
		sym_get(1);
	}
	result->type = optional ? N_optarg : N_arg;

	strcpy(buf, sym_buf);
	parse(S_string);
	strcat(buf, sym_buf);
	parse(S_separator);
	strcat(buf, sym_buf);
	parse(S_string);

	result->value = tac_strdup(buf);
	result->next = parse_attrs();
	return (result);
}


static void
 getsym();

static void
sym_get(int flag)
{
	getsym();
	while (sym_code == S_eol && (flag | SYM_GET_IGNORE_EOL)) {
		getsym();
	}

	if (debug & DEBUG_PARSE_FLAG) {
	report(LOG_DEBUG, "line=%d sym=%s code=%d buf='%s'",
		   sym_line, codestring(sym_code), sym_code, sym_buf);
	}
	if (parse_only && sym_code != S_eof)
		fprintf(stderr, "line %d: symbol %d='%s'\n", sym_line, sym_code, sym_buf);

}

static char *
sym_buf_add(c)
char c;
{
	if (sym_pos >= MAX_INPUT_LINE_LEN) {
	sym_buf[MAX_INPUT_LINE_LEN-1] = '\0';
	if (debug & DEBUG_PARSE_FLAG) {
		report(LOG_DEBUG, "line too long: line=%d sym=%s code=%d buf='%s'",
		   sym_line, codestring(sym_code), sym_code, sym_buf);
	}
	return(NULL);
	}

	sym_buf[sym_pos++] = c;
	return(sym_buf);
}
	
static void
getsym()
{

next:
	switch (sym_ch) {

	case EOF:
		sym_code = S_eof;
		return;

	case '\n':
		sym_line++;
	case '\r':  /* added to parse dos files */
        rch();
		goto next;

	case ';':	/* start looking for end of statement markers */
		sym_code = S_eol;
		strcpy(sym_buf, codestring(S_eol));
		return;

	case '\t':
	case ' ':
		/* eat white space */
		while (sym_ch == ' ' || sym_ch == '\t')
			rch();
		goto next;

	case '=':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_separator;
		rch();
		return;

	case '{':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_openbra;
		rch();
		return;

	case '}':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_closebra;
		rch();
		return;

	case ',':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_comma;
		rch();
		return;

	case '(':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_openparen;
		rch();
		return;

	case ')':
		sym_buf[0] = sym_ch;
		sym_buf[1] = '\0';
		sym_code = S_closeparen;
		rch();
		return;

	case '#':	/* start of comments, eat to end of line */
		while ((sym_ch != '\n') && (sym_ch != EOF))
			rch();
		goto next;

	case '"':
		rch();
		sym_pos = 0;
		while (sym_ch != '"' && sym_ch != EOF) {
	
			/* backslash-double-quote is supported inside strings */
			/* also allow \n */
			if (sym_ch == '\\') {
				rch();
				switch (sym_ch) {
				case '\\':
					/* escape the slash */
				case '"':
					/* escape the quote */
					break;

					/* interpret the following */
				case 'n':
					sym_ch = '\n';
					break;

				case 'r':
					sym_ch = '\r';
					break;

					/* pass on the rest, as is */
				default:
					/* preserve the slash for others */
					if (!sym_buf_add('\\')) {
						sym_code = S_unknown;
						rch();
						return;
					}
					break;
				} /* switch */
			} /* if */
			if (!sym_buf_add(sym_ch)) {
				sym_code = S_unknown;
				rch();
				return;
			} /* if */
			rch();
		} /* while */
		rch();
	
		if (!sym_buf_add('\0')) {
			sym_code = S_unknown;
			rch();
			return;
		}
		sym_code = S_string;
		return;

	default:
	sym_pos = 0;
    while (!strchr("\t\n\r{},()=\";# ", sym_ch)) {

		if (!sym_buf_add(sym_ch)) {
			sym_code = S_unknown;
			rch();
			return;
		}
		rch();
	}

	if (!sym_buf_add('\0')) {
		sym_code = S_unknown;
		rch();
		return;
	}
	sym_code = keycode(sym_buf);
	if (sym_code == S_unknown)
		sym_code = S_string;
	return;
	}
}

static void
rch()
{
	if (sym_error) {
		sym_ch = EOF;
		return;
	}
	sym_ch = getc(cf);
}


/* For a user or group, find the value of a field. Does not recurse. */
VALUE
get_value(user, field)
USER *user;
int field;
{
	VALUE v;

	v.intval = 0;

	if (!user) {
		parse_error("get_value: illegal user");
		return (v);
	}
	switch (field) {

	case S_name:
		v.pval = user->name;
		break;

	case S_login:
		v.pval = user->login;
		break;

	case S_global:
		v.pval = user->global;
		break;

	case S_member:
		v.pval = user->member;
		break;

	case S_expires:
		v.pval = user->expires;
		break;

	case S_arap:
		v.pval = user->arap;
		break;

	case S_chap:
		v.pval = user->chap;
		break;

#ifdef MSCHAP
	case S_mschap:
		v.pval = user->mschap;
		break;
#endif /* MSCHAP */

	case S_pap:
		v.pval = user->pap;
		break;

	case S_opap:
		v.pval = user->opap;
		break;

	case S_message:
		v.pval = user->msg;
		break;

	case S_svc:
		v.pval = user->svcs;
		break;

	case S_before:
		v.pval = user->before_author;
		break;

	case S_after:
		v.pval = user->after_author;
		break;

	case S_svc_dflt:
		v.intval = user->svc_dflt;
		break;

#ifdef MAXSESS
	case S_maxsess:
		v.intval = user->maxsess;
		break;
#endif 

#if 0  /* use auth_method == S_nopassword instead */
	case S_nopasswd:
		v.intval = user->nopasswd;
		break;
#endif
	
	case S_time:
		v.pval = user->time;
		break;

	case S_enable:
		v.pval = user->enable;
		break;

	case S_level:
		v.intval = user->enable_level;
		break;

#ifdef USE_DB
	case S_login_db:
		v.pval = user->db_login;
		break;

	case S_enable_db:
		v.pval = user->db_enable;
		break;

	case S_db_accounting:
		v.pval = user->db_acct;
		break;
#endif
	case S_accounting:
		v.pval = user->acctfile;
		break;

	case S_enable_auth_method:
		v.intval = user->enable_auth_method;
		break;

	case S_login_auth_method:
		v.intval = user->login_auth_method;
		break;

	case S_pap_auth_method:
		v.intval = user->pap_auth_method;
		break;

	default:
		report(LOG_ERR, "get_value: unknown field %d", field);
		break;
	}
	return (v);
}

/* For host , find value of field. Doesn't recursive */
VALUE
get_hvalue(HOST *host, int field)
{
	VALUE v;
	v.intval = 0;
	if(!host) {
		parse_error("get_hvalue: illegal host");
			return (v);
	}
	switch (field) {
	case S_name:
		v.pval = host->name;
		break;
	
	case S_key:
		v.pval = host->key;
		break;

	case S_type:
		v.pval = host->type;
		break;

	case S_prompt:
		v.pval = host->prompt;
		break;
	
	case S_enable:
		v.pval = host->enable;
		break;

	case S_login:
		v.pval = host->login;
		break;

	case S_enable_auth_method:
		v.intval = host->enable_auth_method;
		break;

	case S_login_auth_method:
		v.intval = host->login_auth_method;
		break;

#ifdef USE_DB
	case S_login_db:
		v.pval = host->db_login;
		break;

	case S_enable_db:
		v.pval = host->db_enable;
		break;

	case S_db_accounting:
		v.pval = host->db_acct;
		break;
#endif

	case S_accounting:
		v.pval = host->acctfile;
		break;

	default:
		report(LOG_ERR, "get_value: unknown field %d", field);
		break;
	}
	return (v);
}


/* For each user, check she doesn't circularly reference a
   group. Return 1 if it does */
static int
circularity_check()
{
	USER *user, *entry, *group;
	USER **users = (USER **) hash_get_entries(usertable);
	USER **groups = (USER **) hash_get_entries(grouptable);
	USER **p, **q;

	/* users */
	for (p = users; *p; p++) {
		user = *p;

		if (debug & DEBUG_PARSE_FLAG)
			report(LOG_DEBUG, "circularity_check: user=%s", user->name);

		/* Initialise all groups "seen" flags to zero */
		for (q = groups; *q; q++) {
			group = *q;
			group->flags &= ~FLAG_SEEN;
		}
	
		entry = user;
	
		while (entry) {
			/* check groups we are a member of */
			char *groupname = entry->member;
	
			if (debug & DEBUG_PARSE_FLAG)
			report(LOG_DEBUG, "\tmember of group %s",
				   groupname ? groupname : "<none>");
	
	
			/* if not a member of any groups, go on to next user */
			if (!groupname)
				break;
	
			group = (USER *) hash_lookup(grouptable, groupname);
			if (!group) {
				report(LOG_ERR, "%s=%s, group %s does not exist",
				   (entry->flags & FLAG_ISUSER) ? "user" : "group",
				   entry->name, groupname);
				free(users);
				free(groups);
				return (1);
			}
			if (group->flags & FLAG_SEEN) {
				report(LOG_ERR, "recursively defined groups");
	
			/* print all seen "groups" */
				for (q = groups; *q; q++) {
					group = *q;
					if (group->flags & FLAG_SEEN)
					report(LOG_ERR, "%s", group->name);
				}
				free(users);
				free(groups);
				return (1);
			}
			group->flags |= FLAG_SEEN;	/* mark group as seen */
			entry = group;
		}
	}
	free(users);
	free(groups);
	return (0);
}


/* Return a value for a group or user (isuser says if
   this name is a group or a user name).

   If no value exists, and recurse is true, also check groups we are a
   member of, recursively.

   Returns void * because it can return a string or a node pointer
   (should really return a union pointer).
*/
static VALUE
cfg_get_value(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
	static char		cache_user_name[256] = "";
	static char		cache_group_name[256] = "";
	static USER		*cache_user = NULL, *cache_group;

	USER *user = NULL, *group = NULL;
	VALUE value;

	value.pval = NULL;

	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_value: name=%s isuser=%d attr=%s rec=%d",
		   name, isuser, codestring(attr), recurse);

	/* find the user/group entry */
	if (cache_user_name[0])	{		/* only check if we have a previous hit */
		if (STREQ(name, cache_user_name)) {   /* we have a hit */
			user = cache_user;
			if (debug & DEBUG_CONFIG_FLAG)
				report(LOG_DEBUG, "cfg_get_value: found user name in cache");
		} else {			  /* no hit - clear cache */
		   cache_user_name[0] = '\0';
		   cache_user  = NULL;
		}
	}

	if (!user) {
		user = (USER *) hash_lookup(isuser ? usertable : grouptable, name);
		strcpy(cache_user_name, name);
		cache_user = user;
	}

	if (!user) {  /* user may be from default lookup */
		switch (authen_default_method) {
#ifdef USE_PAM
		case S_pam:
#endif
		case S_unix:
			if (NULL != (name = get_unix_group(name))) {
				isuser = 0;
				user = (USER *) hash_lookup(grouptable, name);
				if (user && debug & DEBUG_CONFIG_FLAG)
					report(LOG_DEBUG, "cfg_get_value: found user in unix group named %s", user->name);
			}
			break;
		} /* switch */
	}

	if (!user) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_value: no user/group named %s", name);
		return (value);
	}

	/* found the entry. Lookup value from attr=value */
	value = get_value(user, attr);

	/* return if we found a value or we are not checking groups because
	   either the search requested or the user does not belong to a group 
	*/
	if (value.pval || !recurse || !user->member) {
		return (value);
	}

	/* JRM Added 6/14/02 cache the first line group */
	if (cache_group_name[0])	{		/* only check if we have a previous hit */
		if (STREQ(user->member, cache_group_name)) {   /* we have a hit */
			group = cache_group;
			if (debug & DEBUG_CONFIG_FLAG)
				report(LOG_DEBUG, "cfg_get_value: found group name in cache");
		} else {			  /* no hit - clear cache */
		   cache_group_name[0] = '\0';
		   cache_group  = NULL;
		}
	}

	if (!group) {
		group = (USER *) hash_lookup(grouptable, user->member);
		strcpy(cache_group_name, user->member);
		cache_group = group;
	}

	while (group) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_value: recurse group = %s",
				   group->name);

		value = get_value(group, attr);

		if (value.pval)
			return (value);

	/* still nothing. Check containing group and so on */

		if (group->member)
			group = (USER *) hash_lookup(grouptable, group->member);
		else
		   group = NULL;
	}

		/* no value for this user or her containing groups */
	value.pval = NULL;
	return (value);
}


/* Wrappers for cfg_get_value */
int
cfg_get_intvalue(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
	int val = cfg_get_value(name, isuser, attr, recurse).intval;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_intvalue: returns %d", val);
	return(val);
}

char *
cfg_get_pvalue(name, isuser, attr, recurse)
char *name;
int isuser, attr, recurse;
{
	char *p = cfg_get_value(name, isuser, attr, recurse).pval;

	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_pvalue: user=%s, attr=%s(%d) returns %s",  
			   name, codestring(attr), attr, p ? p : "NULL");
	return(p);
}

HOST *find_host(tac_uint32 hostaddr)
{
	static tac_uint32	cache_hostaddr = 0;
	static HOST			*cache_host = NULL;
	HOST *idxhost, *host = NULL;
	int idx;

	if (cache_hostaddr)	{		/* only check if we have a previous hit */
		if (hostaddr == cache_hostaddr) {   /* we have a hit */
			host = cache_host;
			if (debug & DEBUG_CONFIG_FLAG)
				report(LOG_DEBUG, "find_host: found host in cache");

		} else {			  /* no hit - clear cache */
		   cache_hostaddr = 0;
		   cache_host  = NULL;
		}
	}

	if (!host) {
		/* find the host entry in host table */
		for (idx=0; idx < cntHost; idx++) {
			idxhost = host_table[idx].host;

			if (debug & DEBUG_CONFIG_FLAG) {
				report(LOG_DEBUG, "find_host: looking for match at index %d, name '%s'",
					   idx+1, idxhost->name);
			}

			if ((hostaddr & host_table[idx].mask) == host_table[idx].addr) {
				if (debug & DEBUG_CONFIG_FLAG) {
					report(LOG_DEBUG, "find_host: found match at index %d, name '%s'",
						   idx+1, idxhost->name);
				}
				host = idxhost;
				break;
			}
		}
		cache_hostaddr = hostaddr;
		cache_host = host;
	}
	return (host);
}

/* For getting host values */
static VALUE
cfg_get_hvalue(tac_uint32 hostaddr, int attr)
{
	HOST *host = NULL;
	VALUE value;

	value.pval = NULL;
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_hvalue: name=%s attr=%s ",
			   cfg_addr_as_string(hostaddr), codestring(attr));

	host = find_host(hostaddr);

	if (host && (debug & DEBUG_CONFIG_FLAG)) {
		report(LOG_DEBUG, "cfg_get_hvalue: found host =%s",
			   host->name);
	}

	if (!host) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_hvalue: no host found for address %s", 
				   cfg_addr_as_string(hostaddr));
		return (value);
	}

    /* found the entry. Lookup value from attr=value */
	return(get_hvalue(host, attr));
}

/* Wrappers for cfg_get_hvalue */
char *
cfg_get_phvalue(tac_uint32 hostaddr, int attr)
{
	char *p = cfg_get_hvalue(hostaddr, attr).pval;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_phvalue: host=%s, attr=%s(%d) returns %s", 
		   cfg_addr_as_string(hostaddr),
		   codestring(attr), attr, p ? p : "NULL");
	return(p);
}

int
cfg_get_inthvalue(tac_uint32 hostaddr, int attr)
{
	int val = cfg_get_hvalue(hostaddr, attr).intval;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_inthvalue: returns %d", val);
	return(val);
}

/*
   Read the config file and do some basic sanity checking on
   it. Return 1 if we find any errors. */
int
cfg_read_config(cfile)
char *cfile;
{
	sym_line = 1;
	sym_error = 0;

	if ((cf = fopen(cfile, "r")) == NULL) {
		report(LOG_ERR, "read_config: fopen() error for file %s %s, exiting",
		   cfile, strerror(errno));
	return (1);
	}
	if (parse_decls() || sym_error) {
		fclose(cf);
		return (1);
	}
	if (parse_only || (debug & DEBUG_PARSE_FLAG)) {
		dump_host_table();
	}
	if (circularity_check()) {
		fclose(cf);
		return (1);
	}

	fclose(cf);
	return (0);
}

/* return 1 if user exists, 0 otherwise */
int
cfg_user_exists(username)
char *username;
{
	USER *user = (USER *) hash_lookup(usertable, username);

	if (!user) {  /* user may be from default lookup */
		switch (authen_default_method) {
#ifdef USE_PAM
		case S_pam:
#endif
		case S_unix:
			return (0 != get_unix_user(username));
			break;
		} /* switch */
	} 

	return (user != NULL);
}

/* returns 1 if user is found in group */
/* no recursion flag because it is implied in search */

int
cfg_is_user_in_group(char *username, char *groupname)
{
	USER *user = NULL, *group = NULL;
	char *name = NULL;

	if (!username || !groupname) {
		/* can't work with NULL values */
		return(0);
	}
	if (!*username || !*groupname) {
		/* or empty strings */
		return(0);
	}

	if (debug & DEBUG_CONFIG_FLAG) {
		report(LOG_DEBUG, "cfg_user_in_group: is user '%s' in group '%s'", 
			   username, groupname);
	}

	if (STREQ(username,groupname)) {
		/* user is considered a member of its own group */
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_user_in_group: user '%s' is a member of '%s'", 
				   username, groupname);
		/* we have a winner, return a yes */
		return(1);
	}

	user = (USER *) hash_lookup(usertable, username);

	if (!user) {  /* user may be from default lookup */
		switch (authen_default_method) {
#ifdef USE_PAM
		case S_pam:
#endif
		case S_unix:
			if (NULL != (name = get_unix_group(username))) {
				if (STREQ(groupname, name)) {
					if (debug & DEBUG_CONFIG_FLAG)
						report(LOG_DEBUG, "cfg_user_in_group: user '%s' is a member of '%s'", 
							   username, groupname);
					/* we have a winner, return a yes */
					return(1);
				} else {
					user = (USER *) hash_lookup(grouptable, name);
				}
			}
			break;
		} /* switch */
	}

	if (!user) {  /* user not found, definitely not in group */
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_user_in_group: no user named '%s'", username);
		return (0);
	}

	/* check first level membership */

	if (!user->member) {
		/* that was quick, this user does not belong to any groups */
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_user_in_group: user '%s' does not belong to group '%s'", 
				   username, groupname);
		return(0);
	}

	if (STREQ(groupname, user->member)) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_user_in_group: user '%s' is a member of '%s'", 
				   username, groupname);
		/* we have a winner, return a yes */
		return(1);
	}

	/* we'll have to keep looking */

	group = (USER *) hash_lookup(grouptable, user->member);

	while (group) {

		if (group->member) {
			if (STREQ(group->name,groupname)) {
				/* stop the presses, the user is in */
				if (debug & DEBUG_CONFIG_FLAG)
					report(LOG_DEBUG, "cfg_user_in_group: user '%s' is a member of '%s'", 
						   username, groupname);
				return(1);
			}
			/* still nothing. Check containing group and so on */

			group = (USER *) hash_lookup(grouptable, group->member);
		} else {
			group = NULL;
		}

		if (group) {
			if (debug & DEBUG_CONFIG_FLAG)
				report(LOG_DEBUG, "cfg_user_in_group: recurse group = '%s'", group->name); 
		}
	}

	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_user_in_group: user '%s' not found in group '%s'", 
		   username, groupname);

	return(0);  /* no dice, return the bad news */
}



/* return expiry string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_expires(char *username, int recurse)
{
	return (cfg_get_pvalue(username, TAC_IS_USER, S_expires, recurse));
}

/* return time string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_timestamp(char *username, int recurse)
{
	return (cfg_get_pvalue(username, TAC_IS_USER, S_time, recurse));
}

char *
cfg_get_user_secret(char *user, int type, int recurse)
{
	int field;
	/* return login information by default, enable if requested */
	switch (type) {
	case S_enable:
		field = S_enable;
		break;

	case S_pap:
		field = S_pap;
		break;

	default:
		field = S_login;
		break;
	}

	return (cfg_get_pvalue(user, TAC_IS_USER, field, recurse));
}


/* return password string of user. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_login_secret(user, recurse)
char *user;
int recurse;
{
	return (cfg_get_user_secret(user, S_login, recurse));
}

char * 
cfg_get_enable_secret(char *user, int recurse)
{
	return (cfg_get_user_secret(user, S_enable, recurse));
}

int
cfg_get_enable_level(char *user, int recurse)
{
	return (cfg_get_intvalue(user, TAC_IS_USER, S_level, recurse));
}

int 
cfg_get_user_auth_method(char *user, int type, int recurse)
{
	int field;
	int auth_method;
	/* return login information by default, enable if requested */
	switch (type) {
	case S_enable:
		field = S_enable_auth_method;
		break;

	case S_pap:
		field = S_pap_auth_method;
		break;

	default:
		field = S_login_auth_method;
		break;
	}
	auth_method = cfg_get_intvalue(user, TAC_IS_USER, field, recurse);

	if (type == S_enable && auth_method == S_login) {  /* if enable = login */
		auth_method = cfg_get_intvalue(user, TAC_IS_USER, S_login_auth_method, recurse);
		if (!auth_method) {
			auth_method = authen_default_method;
		}
	}

	return (auth_method);
}

char *
cfg_get_user_acctfile(char *user, int recurse)
{
	return(cfg_get_pvalue(user, TAC_IS_USER, S_accounting, recurse));
}

char *
cfg_get_user_before_cmd(char *user, int recurse)
{
	return(cfg_get_pvalue(user, TAC_IS_USER, S_before, recurse));
}

char *
cfg_get_user_after_cmd(char *user, int recurse)
{
	return(cfg_get_pvalue(user, TAC_IS_USER, S_after, recurse));
}

#ifdef USE_DB

TAC_DBINFO *
cfg_get_user_dbinfo(char *user, int type, int recurse)
{
	int field;
	TAC_DBINFO	*db_info;

	/* return login information by default, enable if requested */
	switch(type) {
	case S_enable:
		field = S_enable_db;
		break;

	case S_login:
		field = S_login_db;
		break;

	case S_db_accounting:
	case S_accounting:
		field = S_db_accounting;
		break;

	default:
		field = S_login_db;
		break;
	}

	db_info = (TAC_DBINFO *)cfg_get_value(user, TAC_IS_USER, field, recurse).pval;

	if (debug & DEBUG_CONFIG_FLAG) {
		if (db_info) {
			report(LOG_DEBUG, "cfg_get_user_dbinfo: user=%s, attr=%s(%d) returns db_info: " \
				              "db_pref=%s, db_user=%s, " \
					          "db_password=%s, db_hostname=%s, db_name=%s, " \
						      "db_table=%s, dbfield_name=%s, dbfield_passwd=%s", 
		 			user, codestring(field), field,
					db_info->type, db_info->user, db_info->password, 
					db_info->host, db_info->database, db_info->table, 
					db_info->field_name, db_info->field_passwd);
		} else {
			report(LOG_DEBUG, "cfg_get_user_dbinfo: user=%s, attr=%s(%d) returns NULL",
		 			user, codestring(field), field);

		}
	}

	return (db_info);
}

#endif

/* return value of the nopasswd field. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
/* This function has been depreciated.  Use get user auth method instead JRM 6/11/02 */
/*
int
cfg_get_user_nopasswd(user, recurse)
	char *user;
{
	return (cfg_get_intvalue(user, TAC_IS_USER, S_nopasswd, recurse));
}
*/

/* return user's secret. If none, try groups she is a member
   on, and so on, recursively if recurse is non-zero */
char *
cfg_get_arap_secret(user, recurse)
char *user;

{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_arap, recurse));
}

char *
cfg_get_chap_secret(user, recurse)
char *user;

{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_chap, recurse));
}

#ifdef MSCHAP
char *
cfg_get_mschap_secret(user, recurse)
char *user;

{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_mschap, recurse));
}
#endif /* MSCHAP */

char *
cfg_get_pap_secret(user, recurse)
char *user;
{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_pap, recurse));
}

char *
cfg_get_opap_secret(user, recurse)
char *user;
{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_opap, recurse));
}

/* return the global password for the user (or the group, etc.) */

char *
cfg_get_global_secret(user, recurse)
char *user;
int recurse;
{
	return (cfg_get_pvalue(user, TAC_IS_USER, S_global, recurse));
}

#ifdef USE_PAM
/* Return a pointer to a node representing a PAM Service name */
char *
cfg_get_pam_service(user,recurse)
char *user;
int recurse;
{
	char *cfg_passwd;
	int auth_method = 0;
	char *p;   

	cfg_passwd = cfg_get_pap_secret(user, recurse);
	auth_method = cfg_get_user_auth_method(user, S_pap, recurse);
 
	if (!cfg_passwd) {
		cfg_passwd = cfg_get_global_secret(user, recurse);
	}
 
	if (!cfg_passwd && !cfg_user_exists(user)) {
		cfg_passwd = cfg_get_authen_default();

		switch (cfg_get_authen_default_method()) {
		case (S_pam): 
			if (debug & DEBUG_AUTHOR_FLAG)
						report(LOG_DEBUG, "Get default PAM Service :%s",cfg_passwd);
			return(cfg_passwd);
			break;
		default:
			if (debug & DEBUG_AUTHOR_FLAG)
						report(LOG_DEBUG, "No pam service defined");
			return(NULL);/* Haven't any PAM Service!! */
		}
	}

	if (auth_method) {
		p=cfg_passwd;
	} else {
		p=tac_find_substring("pam ", cfg_passwd);
	}

	if(p) {  /* We find PAM services */
		if (debug & DEBUG_AUTHOR_FLAG)
			report(LOG_DEBUG, "PAM service:%s",p);
	return (p);
}

if (debug & DEBUG_AUTHOR_FLAG)
	report(LOG_DEBUG, "No PAM Service");

return(NULL);
}

#endif /* For PAM */
	


/* Return a pointer to a node representing a given service
   authorization, taking care of recursion issues correctly. Protocol
   is only read if the type is N_svc_ppp. svcname is only read if type
   is N_svc.
*/

NODE *
cfg_get_svc_node(name, type, protocol, svcname, recurse)
char *name;
int type;
char *protocol, *svcname;
int recurse;
{
	USER *user, *group;
	NODE *svc;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, 
		   "cfg_get_svc_node: name=%s %s proto=%s svcname=%s rec=%d",
		   name, 
		   cfg_nodestring(type), 
		   protocol ? protocol : "", 
		   svcname ? svcname : "", 
		   recurse);

	/* find the user/group entry */
	user = (USER *) hash_lookup(usertable, name);

	if (!user) {  /* user may be from default lookup */
		switch (authen_default_method) {
#ifdef USE_PAM
		case S_pam:
#endif
		case S_unix:
			if (NULL != (name = get_unix_group(name))) {
				user = (USER *) hash_lookup(grouptable, name);
			}
			break;
		} /* switch */
	}

	if (!user) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_svc_node: no user named %s", name);
		return (NULL);
	}

	/* found the user entry. Find svc node */
	for(svc = (NODE *) get_value(user, S_svc).pval; svc; svc = svc->next) {

	if (svc->type != type) 
		continue;

	if (type == N_svc_ppp && !STREQ(svc->value1, protocol)) {
		continue;
	}

	if (type == N_svc && !STREQ(svc->value1, svcname)) {
		continue;
	}

	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, 
		   "cfg_get_svc_node: found %s proto=%s svcname=%s",
		   cfg_nodestring(type), 
		   protocol ? protocol : "", 
		   svcname ? svcname : "");

	return(svc);
	}

	if (!recurse) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_svc_node: returns NULL");
		return (NULL);
	}

	/* no matching node. Check containing group */
	if (user->member)
	group = (USER *) hash_lookup(grouptable, user->member);
	else
	group = NULL;

	while (group) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_svc_node: recurse group = %s",
				   group->name);

	for(svc = (NODE *) get_value(group, S_svc).pval; svc; svc = svc->next) {

		if (svc->type != type) 
		continue;

		if (type == N_svc_ppp && !STREQ(svc->value1, protocol)) {
		continue;
		}

		if (type == N_svc && !STREQ(svc->value1, svcname)) {
		continue;
		}

		if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, 
			   "cfg_get_svc_node: found %s proto=%s svcname=%s",
			   cfg_nodestring(type), 
			   protocol ? protocol : "", 
			   svcname ? svcname : "");

		return(svc);
	}

	/* still nothing. Check containing group and so on */

	if (group->member)
		group = (USER *) hash_lookup(grouptable, group->member);
	else
		group = NULL;
	}

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_svc_node: returns NULL");

	/* no matching svc node for this user or her containing groups */
	return (NULL);
}

/* Return a pointer to the node representing a set of command regexp
   matches for a user and command, handling recursion issues correctly */
NODE *
cfg_get_cmd_node(name, cmdname, recurse)
char *name, *cmdname;
int recurse;

{
	USER *user, *group;
	NODE *svc;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_cmd_node: name=%s cmdname=%s rec=%d",
		   name, cmdname, recurse);

	/* find the user/group entry */
	user = (USER *) hash_lookup(usertable, name);

	if (!user) {  /* user may be from default lookup */
		switch (authen_default_method) {
#ifdef USE_PAM
		case S_pam:
#endif
		case S_unix:
			if (NULL != (name = get_unix_group(name))) {
				user = (USER *) hash_lookup(grouptable, name);
			}
			break;
		} /* switch */
	}

	if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_cmd_node: no user named %s", name);
	return (NULL);
	}
	/* found the user entry. Find svc node */
	svc = (NODE *) get_value(user, S_svc).pval;

	while (svc) {
	if (svc->type == N_svc_cmd && STREQ(svc->value, cmdname)) {
		if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_cmd_node: found cmd %s %s node",
			   cmdname, cfg_nodestring(svc->type));
		return (svc);
	}
	svc = svc->next;
	}

	if (!recurse) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_cmd_node: returns NULL");
	return (NULL);
	}
	/* no matching node. Check containing group */
	if (user->member)
	group = (USER *) hash_lookup(grouptable, user->member);
	else
	group = NULL;

	while (group) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_get_cmd_node: recurse group = %s",
		   group->name);

	svc = get_value(group, S_svc).pval;

	while (svc) {
		if (svc->type == N_svc_cmd && STREQ(svc->value, cmdname)) {
		if (debug & DEBUG_CONFIG_FLAG)
			report(LOG_DEBUG, "cfg_get_cmd_node: found cmd %s node %s",
			   cmdname, cfg_nodestring(svc->type));
		return (svc);
		}
		svc = svc->next;
	}

	/* still nothing. Check containing group and so on */

	if (group->member)
		group = (USER *) hash_lookup(grouptable, group->member);
	else
		group = NULL;
	}

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_get_cmd_node: returns NULL");

	/* no matching cmd node for this user or her containing groups */
	return (NULL);
}

/* Return an array of character strings representing configured AV
 * pairs, given a username and a service node. 
 *
 * In the AV strings returned, manipulate the separator character to
 * indicate which args are optional and which are mandatory.
 *
 * Lastly, indicate what default permission was configured by setting
 * denyp */

char **
cfg_get_svc_attrs(svcnode, denyp)
NODE *svcnode;
int *denyp;
{
	int i;
	NODE *node;
	char **args;

	*denyp = 1;

	if (!svcnode)
	return (NULL);

	*denyp = (svcnode->dflt == S_deny);

	i = 0;
	for (node = svcnode->value; node; node = node->next)
	i++;

	args = (char **) tac_malloc(sizeof(char *) * (i + 1));

	i = 0;
	for (node = svcnode->value; node; node = node->next) {
	char *arg = tac_strdup(node->value);
	char *p = index(arg, '=');

	if (p && node->type == N_optarg)
		*p = '*';
	args[i++] = arg;
	}
	args[i] = NULL;
	return (args);
}


int
cfg_user_svc_default_is_permit(user)
char *user;

{
	int permit = cfg_get_intvalue(user, TAC_IS_USER, S_svc_dflt,
				   TAC_PLUS_RECURSE);

	switch (permit) {
	default:			/* default is deny */
	case S_deny:
	return (0);
	case S_permit:
	return (1);
	}
}

int
cfg_no_user_permitted()
{
	if (no_user_dflt == S_permit)
		return (1);
	return (0);
}


char *
cfg_get_authen_default()
{
	return (authen_default);
}

/* For describe authentication method(pam,file,db..etc) */
int 
cfg_get_authen_default_method()
{
   return (authen_default_method);
}

#ifdef USE_DB
TAC_DBINFO *
cfg_get_db_authen_default()
{
	if (authen_default_method == S_db) {
		return(db_authen_default);
	}
}
#endif


/* Return 1 if this user has any ppp services configured. Used for
   authorizing ppp/lcp requests */
int
cfg_ppp_is_configured(username, recurse)
	char *username;
	int recurse;
{
	USER *user, *group;
	NODE *svc;

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_ppp_is_configured: username=%s rec=%d",
		   username, recurse);

	/* find the user/group entry */
	user = (USER *) hash_lookup(usertable, username);

	if (!user) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: no user named %s", 
		   username);
	return (0);
	}

	/* found the user entry. Find svc node */
	for(svc = (NODE *) get_value(user, S_svc).pval; svc; svc = svc->next) {

	if (svc->type != N_svc_ppp) 
		continue;

	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: found svc ppp %s node",
		   svc->value1);
	
	return(1);
	}

	if (!recurse) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: returns 0");
	return (0);
	}

	/* no matching node. Check containing group */
	if (user->member)
	group = (USER *) hash_lookup(grouptable, user->member);
	else
	group = NULL;

	while (group) {
	if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: recurse group = %s",
		   group->name);

	for(svc = (NODE *) get_value(group, S_svc).pval; svc; svc = svc->next) {

		if (svc->type != N_svc_ppp)
		continue;

		if (debug & DEBUG_CONFIG_FLAG)
		report(LOG_DEBUG, "cfg_ppp_is_configured: found svc ppp %s node",
			   svc->value1);
	
		return(1);
	}

	/* still nothing. Check containing group and so on */

	if (group->member)
		group = (USER *) hash_lookup(grouptable, group->member);
	else
		group = NULL;
	}

	if (debug & DEBUG_CONFIG_FLAG)
	report(LOG_DEBUG, "cfg_ppp_is_configured: returns 0");

	/* no PPP svc nodes for this user or her containing groups */
	return (0);
}

/* For getting host key */
char *
cfg_get_host_key(tac_uint32 hostaddr)
{
	return (cfg_get_phvalue(hostaddr, S_key));
}

/* For getting host prompt */
char *
cfg_get_host_prompt(tac_uint32 hostaddr)
{
	char *prompt1;

	if (!(prompt1 = (char*)cfg_get_phvalue(hostaddr, S_prompt))) {
		if (!(prompt1=(char*)session.prompt))
			prompt1 = (char*)NULL;
	}
	return (prompt1);
}

/* For getting host enable */
char *
cfg_get_host_enable(tac_uint32 hostaddr)
{
	return (cfg_get_phvalue(hostaddr, S_enable));
}

char *
cfg_get_host_login(tac_uint32 hostaddr)
{
	return (cfg_get_phvalue(hostaddr, S_login));
}

char *
cfg_get_host_secret(tac_uint32 hostaddr, int auth_type)
{
	int field;
	/* return login information by default, enable if requested */
	switch (auth_type) {
	case S_enable:
		field = S_enable;
		break;

	default:
		field = S_login;
		break;
	}

	return (cfg_get_phvalue(hostaddr, field));
}

int
cfg_get_host_auth_method(tac_uint32 hostaddr, int type)
{
	int field;
	/* return login information by default, enable if requested */
	switch (type) {
	case S_enable:
		field = S_enable_auth_method;
		break;

	case S_pap:
		field = S_pap_auth_method;
		break;

	default:
		field = S_login_auth_method;
		break;
	}

	return (cfg_get_inthvalue(hostaddr, field));
}

char *
cfg_get_host_acctfile(tac_uint32 hostaddr)
{
	return(cfg_get_phvalue(hostaddr, S_accounting));
}

#ifdef USE_DB

TAC_DBINFO *
cfg_get_host_dbinfo(tac_uint32 hostaddr, int type)
{
	int field;
	TAC_DBINFO	*db_info;

	switch(type) {
	case S_enable:
		field = S_enable_db;
		break;

	case S_login:
		field = S_login_db;
		break;

	case S_db_accounting:
	case S_accounting:
		field = S_db_accounting;
		break;

	default:
		return(NULL);
		break;
	}

	db_info = (TAC_DBINFO *)cfg_get_hvalue(hostaddr, field).pval;

	if (debug & DEBUG_CONFIG_FLAG) {
		if (db_info) {
			report(LOG_DEBUG, "cfg_get_host_dbinfo: host=%s, attr=%s(%d) returns db_info: " \
				              "db_pref=%s, db_user=%s, " \
					          "db_password=%s, db_hostname=%s, db_name=%s, " \
						      "db_table=%s, dbfield_name=%s, dbfield_passwd=%s", 
		 			cfg_addr_as_string(hostaddr), codestring(field), field,
					db_info->type, db_info->user, db_info->password, 
					db_info->host, db_info->database, db_info->table, 
					db_info->field_name, db_info->field_passwd);
		} else {
			report(LOG_DEBUG, "cfg_get_host_dbinfo: host=%s, attr=%s(%d) returns NULL",
		 			cfg_addr_as_string(hostaddr), codestring(field), field);

		}
	}

	return (db_info);
}

#endif

/* returns:
 if permit_type = permit or deny
   1 if access is permitted
   0 if access is denied
  -2 if parm error
 
   this routine assumes that if enable access is specified, login is allowed.
*/
int 
cfg_check_host_group_access(char *username, tac_uint32 hostaddr, int auth_type, int *pvalue)
{
	int cnt_acl;
	HOST	*host;
	HOST_ACL *hacl;
	char	*name;	
	USER *user = NULL, *group = NULL;
	VALUE value;
	char szAuthType[24];
	char *pFnName = "cfg_check_host_group_access";

	strcpy(szAuthType, codestring(auth_type));

	if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
		report(LOG_DEBUG, "%s: checking %s access to host '%s' for user '%s'",
			   pFnName, szAuthType,  cfg_addr_as_string(hostaddr), username);
	}

	if (auth_type != S_enable && auth_type != S_login) {
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: invalid auth type specified", pFnName );
		}
		return (-2);
	}

	if (pvalue) {
		*pvalue = 0;
	}

	host = find_host(hostaddr);
	if (!host) {
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: access permitted because host not defined", pFnName );
		}

		return (1);	/* permitted if host not defined */
	}

	if (!host->acltable) {
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: access permitted because no host ACL specified", pFnName );
		}

		return (1);		/* permitted if no host acls */
	}
	/* check how many acls we have in this category */

	cnt_acl = host->cnt_acl[auth_type == S_login ? 0 : 1];

	if (!cnt_acl) {
		/* none here, then permit */
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: access permitted because acl not defined", pFnName );
		}
		return (1);		/* permitted if no acl of requested type */
	}

	hacl = (HOST_ACL *) hash_lookup(host->acltable, username);
	name = username;

	if (hacl) {
		/* we have a match on the name, now check the attributes */
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: found acl '%s %s %s'",
				   pFnName, hacl->permit ? "permit" : "deny", szAuthType, name);
		}
		if (hacl->type == auth_type && hacl->permit) {
			if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
				report(LOG_DEBUG, "%s: access permitted by explicit match", pFnName);
			}
			if (pvalue) {
				*pvalue = hacl->value;
			}
			return (1);
		} else {
			/* one more chance to imply permission */
			if (auth_type == S_login && hacl->type == S_enable) {
				/* if an enable is specified for a given user, then login is implied */
				if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
					report(LOG_DEBUG, "%s: access permitted by implicit match", pFnName);
				}
				if (pvalue) {
					*pvalue = hacl->value;
				}
				return (1);
			} 

		}
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: access denied by ACL", pFnName);
		}

		return (0);

	} else {
		if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
			report(LOG_DEBUG, "%s: no acl found for user '%s', checking groups", pFnName, username);
		}

		/* if the acl is not found */
		/* if we were looking for a permit and it was not in the list, then deny */
		/* if we were looking for a deny and it was not in the list, then permit */
		/* but first we have to check the lineage to see if there may be a group */
		user = (USER *) hash_lookup(usertable, username);
		if (!user) { /* user may from default lookup */
			switch (authen_default_method) {
	#ifdef USE_PAM
			case S_pam:
	#endif
			case S_unix:
				if (NULL != (name = get_unix_group(username))) {
					if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
						report(LOG_DEBUG, "%s: looking for ACL in unix group '%s'", pFnName, name);
					}
					if (NULL != (hacl = hash_lookup(host->acltable, name))) {
						/* found a match at the group level */
						if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
							report(LOG_DEBUG, "%s: found acl '%s %s %s'",
								   pFnName, hacl->permit ? "permit" : "deny", szAuthType, name);
							report(LOG_DEBUG, "%s: user '%s' is a member of '%s'", 
								   pFnName, username, name);
						}
						if (hacl->type == auth_type && hacl->permit) {
							if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
								report(LOG_DEBUG, "%s: access permitted by explicit match", pFnName);
							}
							if (pvalue) {
								*pvalue = hacl->value;
							}
							return (1);
						} else {
							if (auth_type == S_login && hacl->type == S_enable) {
								/* if an enable is specified for a given user, then login is implied */
								if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
									report(LOG_DEBUG, "%s: access permitted by implicit match", pFnName);
								}
								if (pvalue) {
									*pvalue = hacl->value;
								}
								return (1);
							} else {
								if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
									report(LOG_DEBUG, "%s: access denied by ACL for '%s'", pFnName, name);
								}

								return (0);
							}
						}
					}
					user = (USER *) hash_lookup(grouptable, name); 

				}
				break;
			}
		}
		if (!user) { 	/* no place else to check */
			if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
				report(LOG_DEBUG, "%s: no match found for %s", pFnName, name);
			}

			return (0);
		}

		if (!user->member) {
			/* no more groups to check */
			if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
				report(LOG_DEBUG, "%s: no more groups for %s", pFnName, name);
			}

			return (0);
		}

		/* we'll have to keep looking */

		group = (USER *) hash_lookup(grouptable, user->member);

		while (group) {
			name = group->name;
			if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
				report(LOG_DEBUG, "%s: recurse group '%s'", pFnName, name);
			}

			if (NULL != (hacl = hash_lookup(host->acltable, name))) {
				if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
					report(LOG_DEBUG, "%s: found acl '%s %s %s'",
						   pFnName, hacl->permit ? "permit" : "deny", szAuthType, name);
				}

				if (hacl->type == auth_type && hacl->permit) {
					if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
						report(LOG_DEBUG, "%s: access permitted by explicit match", pFnName);
					}
					if (pvalue) {
						*pvalue = hacl->value;
					}
					return (1);
				} else {
					if (auth_type == S_login && hacl->type == S_enable) {
						/* if an enable is specified for a given user, then login is implied */
						if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
							report(LOG_DEBUG, "%s: access permitted by implicit match", pFnName);
						}
						if (pvalue) {
							*pvalue = hacl->value;
						}
						return (1);
					} else {
						if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
							report(LOG_DEBUG, "%s: access denied by ACL for '%s'", pFnName, name);
						}
						return (0);
					}
				}
			}
			if (group->member) {
				group = (USER *) hash_lookup(grouptable, group->member);
			} else {
				group = NULL;
			}
		}
	}

	if (debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG)) {
		report(LOG_DEBUG, "%s: nothing else to check so access denied", pFnName);
	}

	return(0);
}



unsigned maskbits[33] = {
	0x00000000, 0x80000000, 0xC0000000, 0xE0000000,
	0xF0000000, 0xF8000000, 0xFC000000, 0xFE000000,
	0xFF000000, 0xFF800000, 0xFFC00000, 0xFFE00000,
	0xFFF00000, 0xFFF80000, 0xFFFC0000, 0xFFFE0000,
	0xFFFF0000, 0xFFFF8000, 0xFFFFC000, 0xFFFFE000,
	0xFFFFF000, 0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00,
	0xFFFFFF00, 0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0,
	0xFFFFFFF0, 0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE,
	0xFFFFFFFF
};

/* return 1 if accepted, 0 if failed parse */

int 
parse_host_name(char *host, tac_uint32 *paddr, tac_uint32 *pmask)
{
	char	szWork[16];  /* working space */
	int		octet[4], mask[4], o = 0, oTmp;
	char	*pInPos, *pOutPos;		 /* position pointers */
	int 	cntDigits;
	int		cntOut;
	tac_uint32	out_addr = 0, out_mask = 0;
	int		cntWild = 0;
	int		bSlashMask = 0;

	bzero(szWork, sizeof(szWork));

	if (!host) {
		return(0);
	}

	pInPos = host;

	while (*pInPos && *pInPos == ' ') {   /* skip leading blanks */
		pInPos++;
	}

	cntDigits = 0;
	pOutPos = szWork;
	cntOut = sizeof(szWork) -1; /* to prevent overflow */

	while (*pInPos && cntOut) {		/* loop to end of string */
	   if (isdigit(*pInPos)) {
		   cntDigits++;
		   *(pOutPos++) = *(pInPos++);
		   cntOut--;
	   } else if ( *pInPos == '.' || *pInPos == '/' ) {
		   if (!cntDigits || bSlashMask || o > 3) {
			   return(0);	/* dot without numbers */	
		   }
		   *pOutPos = '\0';
		   if (szWork[0] == '*') {
			   cntWild++;
			   octet[o] = 0;
			   mask[o] = 255;
			   o++;
		   } else {
			   oTmp = atoi(szWork);
			   if (oTmp > 254) {
				   return(0);	/* invalid octet */
			   }
			   octet[o] = oTmp;	/* store the octet and reset */
			   mask[o] = 0;
			   o++;
		   }
		   
		   if (*pInPos == '/' ) {
			   if (o != 4 || cntWild) {
				   return(0);
			   }
			   bSlashMask++;
		   }

		   pInPos++;
		   cntDigits = 0;
		   pOutPos = szWork;
		   *pOutPos = '\0';
		   cntOut = sizeof(szWork) -1; /* to prevent overflow */

	   } else if (*pInPos == '*') {  /* wildcard */
			if (cntDigits || bSlashMask || o > 3) {
				return(0);	/* can't process digits and wildcard in same octet */
			}
			cntDigits++;
			*(pOutPos++) = *(pInPos++);
			cntOut--;
	   } else {
		   return(0);
	   }
	} /* while */

	if (o < 4) {
		if (cntDigits && o == 3) {	/* this is my last octet */
			*pOutPos = '\0';
			if (szWork[0] == '*') {
				cntWild++;
				octet[o] = 0;
				mask[o] = 255;
			} else {
				oTmp = atoi(szWork);
				if (oTmp > 254) {
					printf("but too high: %d.\n", oTmp);
					return(0);	/* invalid octet */
				}
				octet[o] = oTmp;	/* store the octet */
				mask[o] = 0;
			}
		} else {
			return(0);		/* not enough digits */
		}
	} else if (bSlashMask) {
		*pOutPos = '\0';
		oTmp = atoi(szWork);
		if (oTmp > 32) {
			return(0);
		}
		out_mask = maskbits[oTmp];
		/* for display; */
		mask[0] = (out_mask >> 24) & 0xFF;
		mask[1] = (out_mask >> 16) & 0xFF;
		mask[2] = (out_mask >> 8) & 0xFF;
		mask[3] = out_mask & 0xFF;
	}

#if 0
	printf("IP address: %d.%d.%d.%d\n", octet[0], octet[1], octet[2], octet[3]);
	printf("IP mask: %d.%d.%d.%d\n", mask[0], mask[1], mask[2], mask[3]);
#endif

	out_addr = (octet[0] << 24) | (octet[1] << 16) | (octet[2] << 8) | octet[3];
	if (!out_mask) {
		out_mask = ~((mask[0] << 24) | (mask[1] << 16) | (mask[2] << 8) | mask[3]);
	}
	/* normalize IP address to the mask */
	out_addr &= out_mask;

	*paddr = out_addr;
	*pmask = out_mask;

	return 1;
}

char *
cfg_addr_as_string(tac_uint32 addr)
{
	static	char	szAddr[16];

	unsigned 	o1,o2,o3,o4;

	o1 = (addr >> 24) & 0xFF;
	o2 = (addr >> 16) & 0xFF;
	o3 = (addr >> 8) & 0xFF;
	o4 = addr & 0xFF;

	sprintf(szAddr, "%d.%d.%d.%d", o1, o2, o3, o4);

	return (szAddr);
}
