/* 
   Copyright (c) 1995-1998 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute this software for
   any purpose and without fee is hereby granted, provided that this
   copyright and permission notice appear on all copies of the
   software and supporting documentation, the name of Cisco Systems,
   Inc. not be used in advertising or publicity pertaining to
   distribution of the program without specific prior permission, and
   notice be given in supporting documentation that modification,
   copying and distribution is by permission of Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"
#include "db.h"
#include "expire.h"

/* internal state variables */
#define STATE_AUTHEN_START   0	/* no requests issued */
#define STATE_AUTHEN_GETUSER 1	/* username has been requested */
#define STATE_AUTHEN_GETPASS 2	/* password has been requested */

struct private_data {
    char password[MAX_PASSWD_LEN + 1];
    int state;
};

/* enable_host_verify:
 if host enable authentication exists, attempt to authorize against it
 return 0 if not attempt made,
 return 1 if enable attempted - data->status will be updated
 */


static void
enable(passwd, data)
char *passwd;
struct authen_data *data;
{
    int level = data->NAS_id->priv_lvl;
	char *name = data->NAS_id->username;
	int user_level;
	int auth_method;
	int auth_type;

	if (debug) {       
		report(LOG_INFO, "enable: level %d query for '%s' %s from %s",
			level, name && name[0] ? name : "unknown",
			data->NAS_id->NAS_port && data->NAS_id->NAS_port[0] ?
			data->NAS_id->NAS_port : "unknown",
			session.peer); 
	}

    /* sanity check */
    if (level < TAC_PLUS_PRIV_LVL_MIN || level > TAC_PLUS_PRIV_LVL_MAX) {
		data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
		data->server_msg = tac_strdup("Invalid privilege level in packet");
		report(LOG_ERR, "%s level=%d %s", session.peer, level, data->server_msg);
		return;
    }

#ifdef USE_DB
	if (session.db_default) {
		if ( debug & DEBUG_PASSWD_FLAG )
			report(LOG_DEBUG, "enable: Switching to database verification");
		session.db_default->field_passwd = "enable";
		if (db_verify(data->NAS_id->username, session.peeraddr, data->NAS_id->NAC_address, passwd, session.db_default, TAC_PLUS_RECURSE)) {
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
			return;
		}
	}
#endif

	/* check to see if user is configured with enable parameters */
	user_level = cfg_get_enable_level(name, TAC_PLUS_RECURSE);

	if (user_level) {
		/* there is a level set so compare against requested level */
		if (level > user_level) {
			if (debug) {
				report(LOG_ERR,"enable: user '%s' requested level %d but only allowed %d",
					   name, level, user_level);
			}
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
			data->server_msg = tac_strdup("User not authorized for requested privilege level");
			return;
		}
	}

    /* But firstly check for host enable password */
    
	auth_method = cfg_get_host_auth_method(data->NAS_id->NAS_addr, S_enable);
	if (auth_method) {
		if (!verify(name, passwd, data, TAC_PLUS_RECURSE, S_enable))
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		return;	   /* we return here because once we check the host, we stop */
	}

	/* check if user has a separate authentication method set */
	auth_type = S_enable;
	auth_method = cfg_get_user_auth_method(name, S_enable, TAC_PLUS_RECURSE);
#if 0
	if (auth_method == S_login) {  /* use the same method as the login */
		auth_type = S_login;
		auth_method = cfg_get_user_auth_method(name, S_login, TAC_PLUS_RECURSE);
	}
#endif

	if (auth_method) {
		/* try to validate the user against this method */
		if (debug & DEBUG_AUTHEN_FLAG) {
			report(LOG_DEBUG, "enable: attempting to validate user '%s' with method %s",
				   name, codestring(auth_method));
		}
		if (!verify(name, passwd, data, TAC_PLUS_RECURSE, auth_type))
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		return;
	}

    /* 0 <= level <= 14: look for $enab<n>$ and verify */
    if (level < TAC_PLUS_PRIV_LVL_MAX) {
		char buf[11];

		sprintf(buf, "$enab%d$", level);
		if (!verify(buf, passwd, data, TAC_PLUS_NORECURSE, S_login))
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		return;
    }

    /* 2). level=15. Try $enab15$ or $enable$ (for backwards
       compatibility) and verify */
    

    if (verify("$enable$", passwd, data, TAC_PLUS_NORECURSE, S_login) ||
		verify("$enab15$", passwd, data, TAC_PLUS_NORECURSE, S_login)) {
		return;
    }

    /* return fail */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    return;
}


/*
 * Tacacs enable authentication function. Wants an enable
 * password, and tries to verify it.
 *
 * Choose_authen will ensure that we already have a username before this
 * gets called.
 *
 * We will query for a password and keep it in the method_data.
 *
 * Any strings returned via pointers in authen_data must come from the
 * heap. They will get freed by the caller.
 *
 * Return 0 if data->status is valid, otherwise 1
 */

int
enable_fn(data)
struct authen_data *data;
{
    char *passwd;
    struct private_data *p;
    int pwlen;

    p = (struct private_data *) data->method_data;

    /* An abort has been received. Clean up and return */
    if (data->flags & TAC_PLUS_CONTINUE_FLAG_ABORT) {
		if (data->method_data)
			free(data->method_data);
		data->method_data = NULL;
		return (1);
    }
    /* Initialise method_data if first time through */
    if (!p) {
		p = (struct private_data *) tac_malloc(sizeof(struct private_data));
		bzero(p, sizeof(struct private_data));
		data->method_data = p;
		p->state = STATE_AUTHEN_START;
    }

    /* As we're enabling, we don't need a username, but do we have a
       password? */

    passwd = p->password;

    if (!passwd[0]) {

	/* No password. Either we need to ask for one and expect to get
	 * called again, or we asked but nothing came back, which is fatal */

	switch (p->state) {
		case STATE_AUTHEN_GETPASS:
			/* We already asked for a password. This should be the
				   reply */
			if (data->client_msg) {
			pwlen = MIN((int)strlen(data->client_msg), MAX_PASSWD_LEN);
			} else {
			pwlen = 0;
			}
			strncpy(passwd, data->client_msg, pwlen);
			passwd[pwlen] = '\0';
	    break;

	default:
			/* we may have to check host and level here */
	    /* Request a password */
			{
				char *name = data->NAS_id->username;
				char *hostname = data->NAS_id->NAS_name;
				tac_uint32	hostaddr = data->NAS_id->NAS_addr;
				int level = data->NAS_id->priv_lvl;
				char passbuf[64];
				int auth_method, host_level = 0;

				if (!name) {
					name = "";
				}
				/* check for access based of host acl */
				if (!cfg_check_host_group_access(name, hostaddr, S_enable, &host_level)) {
					sprintf(passbuf, "User '%s' denied enable access to this device\n", name);
					if ( debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG) ) 
						report(LOG_INFO, "enable_fn: denied user '%s' enable access to host %s ",
							   name, hostname );						   
					if (data->server_msg)
						free(data->server_msg);
					if (data->method_data)
						free(data->method_data);
					data->method_data = NULL;
					data->server_msg = tac_strdup(passbuf);
					data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
					log_access(data);
					return(0);
				}

				/* test enable for this user on this host */
				if (host_level) {
					/* there is a level set so compare against requested level */
					if (level > host_level) {
						if ( debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG) ) {
							report(LOG_ERR,"enable_fn: user '%s' requested level %d but only allowed %d on host %s",
								   name, level, host_level, hostname);
						}
						if (data->server_msg)
							free(data->server_msg);
						if (data->method_data)
							free(data->method_data);
						data->method_data = NULL;

						data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
						data->server_msg = tac_strdup("User not authorized for requested privilege level on this host");
						log_access(data);
						return(0);
					}
				}

				/* JRM 6/14/02 --
				   look ahead to see what auth type we might be using and adjust the prompt 
				   if it des or cleartext then is is a host login
				*/
				auth_method = cfg_get_host_auth_method(data->NAS_id->NAS_addr, S_enable);
				if (auth_method) {  /* is there specific host authentication? */
					if (auth_method == S_des || auth_method == S_cleartext) { /* "flat" auth methods */
						strcpy(passbuf, "Host Password: ");
					} else {
						sprintf(passbuf, "%s's host password: ", *name ? name : "unknown");
					}
				} else {
					/* no host specific, now check user specific */
					auth_method = cfg_get_user_auth_method(name, S_enable, TAC_PLUS_RECURSE);
					if (auth_method) {	/* A user specific method exists */
						sprintf(passbuf, "%s's password: ", *name ? name : "unknown");
					} else {
						strcpy(passbuf, "Password: ");
					}

				}

				data->flags = TAC_PLUS_AUTHEN_FLAG_NOECHO;
				data->server_msg = tac_strdup(passbuf);
				data->status = TAC_PLUS_AUTHEN_STATUS_GETPASS;
				p->state = STATE_AUTHEN_GETPASS;
				return (0);
			}
		}
    }

    /* We have a password. Try validating */

    /* Assume the worst */
    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    switch (data->service) {
	case TAC_PLUS_AUTHEN_SVC_ENABLE:
			/* possibly put host and level check here */
			enable(passwd, data);

			log_access(data);

			if (debug) {
				char *name = data->NAS_id->username;

				/* access log here */

				report(LOG_INFO, "enable: query for '%s' %s from %s %s",
				   name && name[0] ? name : "unknown",
				   data->NAS_id->NAS_port && data->NAS_id->NAS_port[0] ?
					   data->NAS_id->NAS_port : "unknown",
				   session.peer, 
				   (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) ?
				   "accepted" : "rejected");
			}
			if (data->status == TAC_PLUS_AUTHEN_STATUS_PASS) {
				if (data->server_msg)
					free(data->server_msg);
				data->server_msg = tac_strdup("\nEnable accepted.");
			}
			break;
		default:
			data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
			report(LOG_ERR, "%s: Bogus service value %d from packet", 
				session.peer, data->service);
			break;
    }

    if (data->method_data)
		free(data->method_data);
    data->method_data = NULL;

    switch (data->status) {
		case TAC_PLUS_AUTHEN_STATUS_ERROR:
		case TAC_PLUS_AUTHEN_STATUS_FAIL:
		case TAC_PLUS_AUTHEN_STATUS_PASS:
			return (0);
		default:
			report(LOG_ERR, "%s: authenticate_fn can't set status %d",
				session.peer, data->status);
			data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
			return (1);
    }
}
