/* 
   Copyright (c) 1995-1998 by Cisco systems, Inc.

   Permission to use, copy, modify, and distribute this software for
   any purpose and without fee is hereby granted, provided that this
   copyright and permission notice appear on all copies of the
   software and supporting documentation, the name of Cisco Systems,
   Inc. not be used in advertising or publicity pertaining to
   distribution of the program without specific prior permission, and
   notice be given in supporting documentation that modification,
   copying and distribution is by permission of Cisco Systems, Inc.

   Cisco Systems, Inc. makes no representations about the suitability
   of this software for any purpose.  THIS SOFTWARE IS PROVIDED ``AS
   IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "tac_plus.h"
#include "expire.h"
#include "time_limit.h"

#include <grp.h>

#ifdef SHADOW_PASSWORDS
#include <shadow.h>
#endif

#ifdef USE_PAM
int
tac_pam_auth(char *UserName,char *Password,struct authen_data *data,char *Service);
#endif /* USE_PAM   */

/* For LDAP verification */
#ifdef USE_LDAP
#include "ldap.h"
#endif /* LDAP */

/* Generic password verification routines for des, file and cleartext
   passwords */

/* Adjust data->status depending on whether a user has expired or not */

static int
passwd_file_verify(char *user, char *supplied_passwd, struct authen_data *data, char *filename);
static int
unix_verify(char *user, char *supplied_passwd, struct authen_data *data);


void
set_expiration_status(exp_date, data)
char *exp_date;
struct authen_data *data;
{
    int expired;

    /* if the status is anything except pass, there's no point proceeding */
    if (data->status != TAC_PLUS_AUTHEN_STATUS_PASS) {
	return;
    }

    /* Check the expiration date, if any. If NULL, this check will return
     * PW_OK */
    expired = check_expiration(exp_date);

    switch (expired) {
		case PW_OK:
			if (debug & DEBUG_PASSWD_FLAG)
				report(LOG_DEBUG, "Password has not expired %s", 
					exp_date ? exp_date : "<no expiry date set>");

			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
			if (data->server_msg)
				free(data->server_msg);
			data->server_msg = tac_strdup("Signon successful.");

		break;

		case PW_EXPIRING:
			if (debug & DEBUG_PASSWD_FLAG)
				report(LOG_DEBUG, "Password will expire soon %s", 
					exp_date ? exp_date : "<no expiry date set>");
			if (data->server_msg)
				free(data->server_msg);
			data->server_msg = tac_strdup("Password will expire soon");
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;

		break;

		case PW_EXPIRED:
			if (debug & DEBUG_PASSWD_FLAG)
				report(LOG_DEBUG, "Password has expired %s", 
				   exp_date ? exp_date : "<no expiry date set>");
			if (data->server_msg)
				free(data->server_msg);
			data->server_msg = tac_strdup("Password has expired");
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
			
		break;

        default:
			report(LOG_ERR, "%s: Bogus return value %d from check_expiration", 
				session.peer, expired);
			data->status = TAC_PLUS_AUTHEN_STATUS_ERROR;
		break;
    }
	return;
}

/* Verify that this user/password is valid.  Works only for cleartext,
   file and des passwords.
   
   Return 1 if password is valid */

int
verify(char *name, char *passwd, struct authen_data *data, int recurse, int auth_type)
{
    char *exp_date;
    char *timestamp;
    char *cfg_login = NULL;
	char *hostname, *clientaddr;
	tac_uint32	hostaddr;
	int auth_method = 0;
	int rc = 0;
#ifdef USE_DB
	TAC_DBINFO	*db_info;
#endif
	char msg[256];

	/* check parms to avoid possible crash */
	if (!name || !passwd || !data) {
		report(LOG_ERR, "verify: called with NULL parameters.");
		return(0);
	}

	/* check auth_type - currently accepts login, enable, and pap */
	if (auth_type != S_login && auth_type != S_enable && auth_type != S_pap) {
		report(LOG_ERR, "verify: called with invalid auth_type of %d", auth_type);
		return(0);
	}

	hostname = data->NAS_id->NAS_name;
	hostaddr = data->NAS_id->NAS_addr;
	clientaddr = data->NAS_id->NAC_address;

	if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		report(LOG_DEBUG,"verify: %s access for user '%s' to port %s on %s from %s",
			   codestring(auth_type), name, 
			   data->NAS_id->NAS_port, hostname,  
			   clientaddr);
/*
#ifdef USE_DB
	if (session.db_default) {
		report(LOG_INFO,"verify: authenticating via default db");
		if (db_verify(name, session.peeraddr, clientaddr, passwd, session.db_default, recurse)) {
//		if (db_verify(name, (char *)NULL, passwd, session.db_default, recurse)) {
	            data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
		    return (1);
       		 }
	}
#endif	
*/
	/* check host permissions */
	if (auth_type == S_login || auth_type == S_enable) {  /* only check host for shell */
		/* first check for deny */
		if (!cfg_check_host_group_access(name, hostaddr, auth_type, NULL)) {
			sprintf(msg, "User '%s' denied %s access to this device",
				name, codestring(auth_type));
			if ( debug & (DEBUG_AUTHEN_FLAG|DEBUG_PASSWD_FLAG) ) 
				report(LOG_INFO, "verify: denied %s %s access to %s ",
					   name, codestring(auth_type), hostname );						   
			if (data->server_msg)
				free(data->server_msg);
			data->server_msg = tac_strdup(msg);
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
			return(0);
		}
	}

	if (auth_type != S_enable) {  /* we do not need to double check this for enable */
		timestamp = (char *)cfg_get_timestamp(name, recurse); 
		if ( timestamp != NULL ) { 
			if( time_limit_process(timestamp) == 0  ) {
				if ( debug & DEBUG_AUTHEN_FLAG ) 
					report(LOG_DEBUG,"Timestamp check failed");	
				data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
				return (0);
			} 
		}
	}
	
	if (auth_type == S_login || auth_type == S_enable) {  /* only check host for shell */
		/* first order: host specific login */
		auth_method = cfg_get_host_auth_method(hostaddr, auth_type);
		if (auth_method) {	/* if there is not an auth method, no need to look further */
			cfg_login  = cfg_get_host_secret(hostaddr, auth_type);
#ifdef USE_DB
			db_info    = cfg_get_host_dbinfo(hostaddr, auth_type);
#endif
			if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG)) {
				report(LOG_DEBUG, "verify: using host auth parameters");
			}
		}
	}

	if (!auth_method) {
		/* second order: user/group specific */
		auth_method = cfg_get_user_auth_method(name, auth_type, recurse);
		if (auth_method) {	/* if there is not an auth method, no need to look further */
			cfg_login  = cfg_get_user_secret(name, auth_type, recurse);
#ifdef USE_DB
			db_info    = cfg_get_user_dbinfo(name, auth_type, recurse);
#endif
			if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG)) {
				report(LOG_DEBUG, "verify: using user/group auth parameters");
			}
		}
	}

    /* If there is no login or pap password for this user, see if there is 
       a global password for her that we can use */

    if (!auth_method && auth_type != S_enable) {	/* More on this one later  JRM */
		char *pTmp;
		if (NULL != (cfg_login = cfg_get_global_secret(name, recurse))) {
			if (NULL != (pTmp = tac_find_substring("cleartext ", cfg_login))) {
				auth_method = S_cleartext;
				if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG)) {
					report(LOG_DEBUG, "verify: using global auth parameters");
				}
			}
			cfg_login = pTmp;
		}
    }

    /* If we still have no password for this user (or no user for that
       matter) look for default authentication = <login_spec>  */

	if (!auth_method && auth_type != S_enable) {	/* only lookup default for login or pap */
		auth_method = cfg_get_authen_default_method();
		if (auth_method) {	/* if there is not an auth method, no need to look further */
			cfg_login = cfg_get_authen_default();
	#ifdef USE_DB
			db_info = cfg_get_db_authen_default();
	#endif
			if (debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG)) {
				report(LOG_DEBUG, "verify: using default auth parameters");
			}
		}
	}

	/* If we do not know what we are doing at this point, its time to bail */

	if (!auth_method) {
		report(LOG_ERR, "verify: failed - could not authenticate"
				" for user '%s' on NAS '%s'", name, hostname);
		return(0);
	}

    /* We have an auth_method and a configured password. Deal with it depending on its
       type */

	if (debug & DEBUG_PASSWD_FLAG)
		report(LOG_DEBUG, "verify: Using auth_method %s(%d) with data %s", 
			   codestring(auth_method), auth_method, cfg_login);

	if (auth_type != S_enable) {
		exp_date = cfg_get_expires(name, recurse);
	}

	switch (auth_method) {
	
	case S_cleartext:

		if (STREQ(passwd, cfg_login)) {
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
		} else {
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		}
		break;

#ifdef USE_LDAP
	case (S_ldap):
		if (ldap_verify(name, passwd, cfg_login)==1) {
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		} else {
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
			exp_date = NULL;	/* no expire check for LDAP */
		}
		break;
#endif /* USE_LDAP */

#ifdef USE_PAM
	case S_pam:

		if (tac_pam_auth(name, passwd, data, cfg_login)) {
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		} else {
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
		}
		break;

#endif /* USE_PAM */

	case S_des:
		/* try to verify this des password */
		if (!des_verify(passwd, cfg_login)) {
			data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		} else {
			data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
		}
		break;

#ifdef USE_DB
	case S_db:
        /* try to verify this password from database */
		if (!db_verify(name, session.peeraddr, clientaddr, passwd, db_info, recurse)) {
//		if (!db_verify(name, (char*)NULL, passwd, db_info, recurse)) {
	            data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
       		 } else {
	            data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
       		 }
		break;

#endif /* USE_DB */
	case S_file:
		return (passwd_file_verify(name, passwd, data, cfg_login));

	case S_unix:
		return (unix_verify(name, passwd, data));

	default:
		report(LOG_ERR, "enable_host_verify: unsupported enable method %s(%d)", 
				   codestring(auth_method), auth_method);
		data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
		return(0);

	} /* switch */
    
	if (auth_type != S_enable) {
		set_expiration_status(exp_date, data);
	}

	rc = (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);

	if (debug & DEBUG_PASSWD_FLAG)
		report(LOG_DEBUG, "verify: %s %s authentication %ssuccessful", 
			   codestring(auth_type), codestring(auth_method), 
			   rc ? "" : "un");

	return(rc);
}

/* verify that this user/password is valid per /etc/passwd.
   Return 0 if invalid. */
static int
unix_verify(char *user, char *supplied_passwd, struct authen_data *data)
{
    struct passwd *pw;
    char *exp_date;
    char *cfg_passwd;
#ifdef SHADOW_PASSWORDS
    char buf[12];
#endif /* SHADOW_PASSWORDS */

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

    setpwent();
    pw = getpwnam(user);
    endpwent();

    if (pw == NULL) {
	/* no entry exists */
	return (0);
    }

    if (*pw->pw_passwd == '\0' ||
	supplied_passwd == NULL ||
	*supplied_passwd == '\0') {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    cfg_passwd = pw->pw_passwd;
    exp_date = pw->pw_shell;

#ifdef SHADOW_PASSWORDS
    if (STREQ(pw->pw_passwd, "x")) {
	struct spwd *spwd = getspnam(user);

	if (!spwd) {
	    if (debug & DEBUG_PASSWD_FLAG) {
		report(LOG_DEBUG, "unix_verify: No entry for %s in shadow file", user);
	    }
	    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	    return (0);
	}
	if (debug & DEBUG_PASSWD_FLAG) {
	    report(LOG_DEBUG, "unix_verify: Found entry for %s in shadow file", user);
	}
	cfg_passwd = spwd->sp_pwdp;

	/* 
	 * Sigh. The Solaris shadow password file contains its own
	 * expiry date as the number of days after the epoch
	 * (January 1, 1970) when the password expires.
	 * Convert this to ascii so that the traditional tacacs
	 * password expiration routines work correctly. 
	 */

	if (spwd->sp_expire > 0) {
	    long secs = spwd->sp_expire * 24 * 60 * 60;
	    char *p = ctime(&secs);
	    bcopy(p+4, buf, 7);
	    bcopy(p+20, buf+7, 4);
	    buf[11] = '\0';
	    exp_date = buf;
	}
    }
#endif /* SHADOW_PASSWORDS */

    /* try to verify the password */
    if (!des_verify(supplied_passwd, cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

    /* password ok. Check expiry field */
    set_expiration_status(exp_date, data);

    return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}

char *get_unix_user(char *user)
{
    struct passwd *pw;

	if (debug & (DEBUG_PASSWD_FLAG|DEBUG_CONFIG_FLAG)) {
		report(LOG_DEBUG, "get_unix_user: user=%s", user);
	}
	if (!user) {
		return (NULL);  /* can't lookup null user */
	}

	pw = getpwnam(user);

	if (pw) {
		if (debug & (DEBUG_PASSWD_FLAG|DEBUG_CONFIG_FLAG)) {
			report(LOG_DEBUG, "get_unix_user: found user=%s, uid=%u, gid=%u", 
				   user, pw->pw_uid, pw->pw_gid);
		}
		return (pw->pw_name);
	}

	return (NULL);
}


char *get_unix_group(char *user)
{
    struct passwd *pw;
	struct group  *grp;

	if (debug & (DEBUG_PASSWD_FLAG|DEBUG_CONFIG_FLAG)) {
		report(LOG_DEBUG, "get_unix_group: user=%s", user);
	}
	if (!user) {
		return (NULL);  /* can't lookup null user */
	}

	pw = getpwnam(user);

	if (pw) {
		if (debug & (DEBUG_PASSWD_FLAG|DEBUG_CONFIG_FLAG)) {
			report(LOG_DEBUG, "get_unix_group: found user=%s, uid=%u, gid=%u", 
				   user, pw->pw_uid, pw->pw_gid);
		}

		grp = getgrgid(pw->pw_gid);
		if (grp) {
			if (debug & (DEBUG_PASSWD_FLAG|DEBUG_CONFIG_FLAG)) {
				report(LOG_DEBUG, "get_unix_group: user=%s is in group %s", 
					   pw->pw_name, grp->gr_name); 
			}
			return (grp->gr_name);
		}
	}
	return(NULL);  /* nothing found */

}


/* verify that this user/password is valid per a passwd(5) style
   database. Return 0 if invalid. */

static int
passwd_file_verify(char *user, char *supplied_passwd, struct authen_data *data, char *filename)
{
    struct passwd *pw;
    char *exp_date;
    char *cfg_passwd;

    data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;

#if 0
    if (filename && (STREQ(filename, "/etc/passwd")|| STREQ(filename,"/etc/shadow") )) {
	return(etc_passwd_file_verify(user, supplied_passwd, data));
    }
#endif

    /* an alternate filename */
    if (!(access(filename, R_OK) == 0)) {
	report(LOG_ERR, "passwd_file_verify: %s %s: Cannot access %s for user %s -- %s",
	       session.peer, session.port, filename, user, strerror(errno));
	return (0);
    }

    pw = tac_passwd_lookup(user, filename);

    if (pw == NULL)
	/* no entry exists */
	return (0);

    if (*pw->pw_passwd == '\0' ||
	supplied_passwd == NULL ||
	*supplied_passwd == '\0') {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    }
    cfg_passwd = pw->pw_passwd;
    exp_date = pw->pw_shell;

    /* try to verify the password */
    if (!des_verify(supplied_passwd, cfg_passwd)) {
	data->status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	return (0);
    } else {
	data->status = TAC_PLUS_AUTHEN_STATUS_PASS;
    }

    /* password ok. Check expiry field */
    set_expiration_status(exp_date, data);
    return (data->status == TAC_PLUS_AUTHEN_STATUS_PASS);
}

/*
 * verify a provided password against a des encrypted one
 * return 1 if verified, 0 otherwise.
 */

int
des_verify(users_passwd, encrypted_passwd)
char *users_passwd, *encrypted_passwd;
{
    char *ep;

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "des_verify: %s", encrypted_passwd);

    if (users_passwd == NULL ||
	*users_passwd == '\0' ||
	encrypted_passwd == NULL ||
	*encrypted_passwd == '\0') {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "des_verify: returns 0");
	return (0);
    }

    ep = (char *) crypt(users_passwd, encrypted_passwd);

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "des_verify: %s encrypts to %s", users_passwd, ep);

    if (STREQ(ep, encrypted_passwd)) {
	if (debug & DEBUG_PASSWD_FLAG)
	    report(LOG_DEBUG, "des_verify: Passwords match");
	return (1);
    }

    if (debug & DEBUG_PASSWD_FLAG)
	report(LOG_DEBUG, "des_verify: Passwords do not match");

    return (0);
}
